import asyncio
import hashlib
import json
import logging
from datetime import datetime
from functools import partial

from pymobiledevice3.services.web_protocol.cdp_screencast import ScreenCast

logger = logging.getLogger(__name__)

NETWORK_RESOURCE_TYPES = [
    "Document",
    "Stylesheet",
    "Image",
    "Media",
    "Font",
    "Script",
    "TextTrack",
    "XHR",
    "Fetch",
    "EventSource",
    "WebSocket",
    "Manifest",
    "SignedExchange",
    "Ping",
    "CSPViolationReport",
    "Preflight",
    "Other",
]

LOG_MESSAGE_SOURCES = {
    "xml": "xml",
    "javascript": "javascript",
    "network": "network",
    "console-api": "javascript",
    "storage": "storage",
    "appcache": "appcache",
    "rendering": "rendering",
    "css": "rendering",
    "security": "security",
    "deprecation": "deprecation",
    "worker": "worker",
    "violation": "violation",
    "intervention": "intervention",
    "recommendation": "recommendation",
    "other": "other",
    "content-blocker": "other",
    "media": "other",
    "mediasource": "other",
    "webrtc": "other",
    "itp-debug": "other",
    "ad-click-attribution": "other",
}

LOG_MESSAGE_LEVELS = {
    "log": "info",
    "info": "info",
    "warning": "warning",
    "error": "error",
    "debug": "verbose",
}

DEBUGGER_PAUSED_REASON = {
    "XHR": "XHR",
    "Fetch": "other",
    "DOM": "DOM",
    "AnimationFrame": "other",
    "Interval": "other",
    "Listener": "EventListener",
    "Timeout": "other",
    "exception": "exception",
    "assert": "assert",
    "CSPViolation": "CSPViolation",
    "DebuggerStatement": "debugCommand",
    "Breakpoint": "instrumentation",
    "PauseOnNextStatement": "instrumentation",
    "Microtask": "other",
    "BlackboxedScript": "other",
    "other": "other",
}


class CdpTarget:
    def __init__(self, protocol, target_id: str):
        """
        :param pymobiledevice3.services.web_protocol.session_protocol.SessionProtocol protocol: Session protocol.
        :param target_id: Target id.
        """
        self.protocol = protocol
        self.target_id = target_id
        self.frame = {}
        self.session_id = protocol.id_
        self.app_id = protocol.app.id_
        self.page_id = protocol.page.id_
        self.output_queue = asyncio.Queue()
        self.input_queue = asyncio.Queue()
        self.screencast = None
        self.from_cdp_special_messages_methods = {
            "Audits.enable": self._audits_enable,
            "DOM.getBoxModel": self._dom_get_box_model,
            "DOM.enable": partial(self._simple_response, value=None),
            "DOM.getNodeForLocation": self._dom_get_node_for_location,
            "DOM.getNodesForSubtreeByStyle": self._dom_get_nodes_for_subtree_by_style,
            "Log.clear": self._log_clear,
            "Log.disable": self._log_disable,
            "Log.enable": self._log_enable,
            "Log.startViolationsReport": partial(self._simple_response, value=None),
            "Page.getNavigationHistory": self._page_get_navigation_history,
            "Page.startScreencast": self._page_start_screencast,
            "Page.stopScreencast": self._page_stop_screencast,
            "Page.screencastFrameAck": self._page_screencast_frame_ack,
            "Page.getResourceTree": self._page_get_resource_tree,
            "Emulation.setEmulatedMedia": self._emulation_set_emulated_media,
            "Emulation.setTouchEmulationEnabled": partial(self._simple_response, value=None),
            "Emulation.setFocusEmulationEnabled": partial(self._simple_response, value=None),
            "Emulation.setEmulatedVisionDeficiency": partial(self._simple_response, value=None),
            "Emulation.setAutoDarkModeOverride": self._emulation_set_auto_dark_mode_override,
            "Emulation.setEmitTouchEventsForMouse": partial(self._simple_response, value=None),
            "Debugger.setAsyncCallStackDepth": partial(self._simple_response, value=True),
            "Debugger.setBlackboxPatterns": self._debugger_set_blackbox_patterns,
            "Debugger.setBreakpointByUrl": self._debugger_set_breakpoint_by_url,
            "DOMDebugger.setBreakOnCSPViolation": partial(self._simple_response, value=None),
            "DOMDebugger.getEventListeners": self._domdebugger_get_event_listeners,
            "Network.setCacheDisabled": self._network_set_cache_disabled,
            "Network.loadNetworkResource": self._network_load_network_resource,
            "Network.setAttachDebugStack": partial(self._simple_response, value=None),
            "Network.clearAcceptedEncodingsOverride": partial(self._simple_response, value=None),
            "ServiceWorker.enable": self._service_worker_enable,
            "HeapProfiler.enable": partial(self._simple_response, value=None),
            "Overlay.setShowGridOverlays": partial(self._simple_response, value=None),
            "Overlay.setShowFlexOverlays": partial(self._simple_response, value=None),
            "Overlay.setShowScrollSnapOverlays": partial(self._simple_response, value=None),
            "Overlay.setShowContainerQueryOverlays": partial(self._simple_response, value=None),
            "Overlay.setShowIsolatedElements": partial(self._simple_response, value=None),
            "Overlay.hideHighlight": partial(self._simple_response, value=None),
            "Overlay.highlightNode": self._overlay_highlight_node,
            "Overlay.enable": partial(self._simple_response, value=None),
            "Overlay.setShowViewportSizeOnResize": partial(self._simple_response, value=None),
            "Overlay.setPausedInDebuggerMessage": partial(self._simple_response, value=None),
            "Runtime.runIfWaitingForDebugger": partial(self._simple_response, value=None),
            "Runtime.compileScript": self._runtime_compile_script,
            "Runtime.getIsolateId": self._runtime_get_isolate_id,
            "Profiler.enable": partial(self._simple_response, value=None),
            "Target.setAutoAttach": self._target_set_auto_attach,
            "Target.setDiscoverTargets": partial(self._simple_response, value=None),
            "Target.setRemoteLocations": partial(self._simple_response, value=None),
            "CSS.trackComputedStyleUpdates": partial(self._simple_response, value=None),
            "CSS.takeComputedStyleUpdates": self._css_take_computed_style_updates,
            "CSS.addRule": self._css_add_rule,
            "Input.emulateTouchFromMouseEvent": self._input_emulate_touch_from_mouse_event,
            "Input.dispatchKeyEvent": self._input_dispatch_key_event,
        }
        self.to_cdp_special_messages_methods = {
            "Target.targetCreated": self._target_created,
            "Target.targetDestroyed": self._target_destroyed,
            "Target.dispatchMessageFromTarget": self._target_dispatch_message_from_target,
            "Target.didCommitProvisionalTarget": self._target_did_commit_provisional_target,
        }
        self.to_cdp_special_dispatched_messages_methods = {
            "Debugger.scriptParsed": self._debugger_script_parsed,
            "Debugger.scriptFailedToParse": self._debugger_script_failed_to_parse,
            "Debugger.paused": self._debugger_paused,
            "Debugger.globalObjectCleared": self._debugger_global_object_cleared,
            "Page.defaultAppearanceDidChange": self._page_default_appearance_did_change,
            "Runtime.executionContextCreated": self._runtime_execution_context_created,
            "Console.messageAdded": self._console_message_added,
            "Network.responseReceived": self._network_response_received,
            "Network.loadingFinished": self._network_loading_finished,
        }
        self._waiting_for_id = 0
        self._input_task = asyncio.create_task(self._input_loop())
        self._receiving_task = asyncio.create_task(self._receive_loop())
        self._script_source_to_context_id = {}
        self._default_execution_id = 0

    @classmethod
    async def create(cls, protocol):
        """
        :param pymobiledevice3.services.web_protocol.session_protocol.SessionProtocol protocol: Session protocol.
        """
        await protocol.inspector.setup_inspector_socket(protocol.id_, protocol.app.id_, protocol.page.id_)
        while not protocol.inspector.wir_events:
            await asyncio.sleep(0)
        created = protocol.inspector.wir_events.pop(0)
        while "targetInfo" not in created["params"]:
            created = protocol.inspector.wir_events.pop(0)
        target_id = created["params"]["targetInfo"]["targetId"]
        logger.info(f"Created: {target_id}")
        target = cls(protocol, target_id)
        await target.output_queue.put(created)
        return target

    async def send(self, message):
        """
        Send message from devtools to the target.
        """
        await self.input_queue.put(message)

    async def receive(self):
        """
        Get message from the target to the devtools.
        """
        return await self.output_queue.get()

    async def wait_for_event_id(self, id_: int):
        """
        Wait for a message with a specific id from the target.
        :param id_: Message id to wait for.
        """
        while True:
            for i in range(len(self.protocol.inspector.wir_events)):
                message = self.protocol.inspector.wir_events[i]
                if message["method"] != "Target.dispatchMessageFromTarget":
                    continue
                message = json.loads(message["params"]["message"])
                if message.get("id", "") != id_:
                    continue
                del self.protocol.inspector.wir_events[i]
                return message
            await asyncio.sleep(0)

    async def send_message_with_result(self, id_, method, params):
        """
        Send a message to the target and wait for response.
        """
        self._waiting_for_id += 1
        await self._send_message_to_target({"id": id_, "method": method, "params": params})
        result = await self.wait_for_event_id(id_)
        self._waiting_for_id -= 1
        return result

    async def evaluate_and_result(self, id_, expression):
        """
        Evaluate Javascript expression.
        """
        logger.debug("Evaluating: ", expression)
        params = {"expression": expression}
        data = await self.send_message_with_result(id_, "Runtime.evaluate", params)
        logger.debug("Evaluated: ", data)
        result = data["result"]["result"]
        if result["type"] == "string":
            return result["value"]
        elif result["type"] == "undefined":
            return None
        elif result["type"] == "object":
            return result
        else:
            logger.debug("Unknown type: ", result)
            return result

    async def _input_loop(self):
        while True:
            message = await self.input_queue.get()
            if message["method"] in self.from_cdp_special_messages_methods:
                await self.from_cdp_special_messages_methods[message["method"]](message)
            else:
                await self._send_message_to_target(message)

    async def _receive_loop(self):
        while True:
            if self._waiting_for_id or not self.protocol.inspector.wir_events:
                await asyncio.sleep(0)
                continue
            message = self.protocol.inspector.wir_events.pop(0)
            await self._to_output_queue(message)

    async def _to_output_queue(self, message):
        if message["method"] in self.to_cdp_special_messages_methods:
            await self.to_cdp_special_messages_methods[message["method"]](message)
        else:
            logger.error("Error!!!!!!!!!!!!", message)
            raise RuntimeError()

    async def _send_message_to_target(self, message):
        await self.protocol.send_command(
            "Target.sendMessageToTarget", targetId=self.target_id, message=json.dumps(message)
        )

    async def _simple_response(self, message, value):
        await self.output_queue.put({"id": message["id"], "result": {"result": value}})

    async def _audits_enable(self, message):
        message["method"] = "Audit.setup"
        await self._send_message_to_target(message)

    async def _dom_get_box_model(self, message):
        message["method"] = "DOM.highlightNode"
        message["params"]["highlightConfig"] = {
            "showInfo": True,
            "contentColor": {"r": 111, "g": 168, "b": 220, "a": 0.66},
            "paddingColor": {"r": 147, "g": 196, "b": 125, "a": 0.55},
            "borderColor": {"r": 255, "g": 229, "b": 153, "a": 0.66},
            "marginColor": {"r": 246, "g": 178, "b": 107, "a": 0.66},
        }
        await self._send_message_to_target(message)

    async def object_id_to_node_id(self, object_id, id_):
        node = await self.send_message_with_result(id_, "DOM.requestNode", {"objectId": object_id})
        return node["result"]["nodeId"]

    async def _dom_get_node_for_location(self, message):
        x, y = message["params"]["x"], message["params"]["y"]
        obj = await self.evaluate_and_result(message["id"], f"document.elementFromPoint({x},{y})")
        if "objectId" not in obj:
            await self._simple_response(message, None)
            return
        result = {"nodeId": await self.object_id_to_node_id(obj["objectId"], message["id"])}
        await self.output_queue.put({"id": message["id"], "result": result})

    async def _dom_get_nodes_for_subtree_by_style(self, message):
        object_id = (
            await self.send_message_with_result(
                message["id"], "DOM.resolveNode", {"nodeId": message["params"]["nodeId"]}
            )
        )["result"]["object"]["objectId"]
        result = await self.send_message_with_result(
            message["id"],
            "Runtime.callFunctionOn",
            {
                "objectId": object_id,
                "functionDeclaration": (
                    "function(styles) {"
                    "   const result = new Set();"
                    '   var all = this.getElementsByTagName("*");'
                    "   for (var elem_i=0; elem_i < all.length; elem_i++) {"
                    "       for (var style_i in styles) {"
                    "           if (window.getComputedStyle(all[elem_i]).getPropertyValue(styles[style_i].name) ==="
                    " styles[style_i].value) {"
                    "               result.add(all[elem_i]);"
                    "               break;"
                    "           }"
                    "       }"
                    "   }"
                    "   return result;"
                    "}"
                ),
                "arguments": [{"value": message["params"]["computedStyles"]}],
            },
        )
        result = await self.send_message_with_result(
            message["id"], "Runtime.getCollectionEntries", {"objectId": result["result"]["result"]["objectId"]}
        )
        nodes = await asyncio.gather(
            *[
                self.object_id_to_node_id(obj["value"]["objectId"], message["id"] - i)
                for i, obj in enumerate(result["result"]["entries"])
            ],
            return_exceptions=True,
        )
        nodes = [n for n in nodes if isinstance(n, int)]
        await self.output_queue.put({"id": message["id"], "result": {"nodeIds": nodes}})

    async def _log_clear(self, message):
        message["method"] = "Console.clearMessages"
        await self._send_message_to_target(message)

    async def _log_disable(self, message):
        message["method"] = "Console.disable"
        await self._send_message_to_target(message)

    async def _log_enable(self, message):
        message["method"] = "Console.enable"
        await self._send_message_to_target(message)

    async def _page_get_navigation_history(self, message):
        href = await self.evaluate_and_result(message["id"], "window.location.href")
        title = await self.evaluate_and_result(message["id"], "document.title")
        await self.output_queue.put({
            "id": message["id"],
            "result": {"currentIndex": 0, "entries": [{"id": 0, "url": href, "title": title}]},
        })

    async def _page_start_screencast(self, message):
        params = message["params"]
        self.screencast = ScreenCast(self, params["format"], params["quality"], params["maxWidth"], params["maxHeight"])
        await self.screencast.start(message["id"])
        await self._simple_response(message, None)

    async def _page_stop_screencast(self, message):
        if self.screencast is not None:
            await self.screencast.clear()
            self.screencast = None
        await self._simple_response(message, None)

    async def _page_screencast_frame_ack(self, message):
        if self.screencast is not None:
            self.screencast.ack(message["params"]["sessionId"])
        await self._simple_response(message, None)

    async def _page_get_resource_tree(self, message):
        result = await self.send_message_with_result(message["id"], message["method"], message["params"])
        self.frame = result["result"]["frameTree"]["frame"]
        await self.output_queue.put(result)

    async def _emulation_set_emulated_media(self, message):
        message["method"] = "Page.setEmulatedMedia"
        await self._send_message_to_target(message)

    async def _emulation_set_auto_dark_mode_override(self, message):
        message["method"] = "Page.setForcedAppearance"
        params = message["params"]
        if not params:
            await self._simple_response(message, None)
            return
        message["params"] = {"appearance": "Dark" if params["enabled"] else "Light"}
        await self._send_message_to_target(message)

    async def _debugger_set_blackbox_patterns(self, message):
        for pattern in message["params"]["patterns"]:
            await self.send_message_with_result(
                message["id"], "Debugger.setShouldBlackboxURL", {"url": pattern, "shouldBlackbox": True}
            )
        await self._simple_response(message, None)

    async def _debugger_set_breakpoint_by_url(self, message):
        condition = message["params"].pop("condition", "")
        if condition:
            message["params"]["options"]["condition"] = condition
        await self._send_message_to_target(message)

    async def _domdebugger_get_event_listeners(self, message):
        node = {"nodeId": await self.object_id_to_node_id(message["params"]["objectId"], message["id"])}
        listeners = await self.send_message_with_result(message["id"], "DOM.getEventListenersForNode", node)
        if "error" in listeners:
            await self._simple_response(message, None)
            return
        listeners_out = []
        for listener in listeners["result"]["listeners"]:
            data = {
                "type": listener["type"],
                "useCapture": listener["useCapture"],
                "passive": listener.get("passive", False),
                "once": listener.get("once", False),
            }
            if "location" in listener:
                data["scriptId"] = listener["location"]["scriptId"]
                data["lineNumber"] = listener["location"]["lineNumber"]
                data["columnNumber"] = listener["location"]["columnNumber"]
            listeners_out.append(data)
        await self.output_queue.put({"id": message["id"], "result": {"listeners": listeners_out}})

    async def _network_set_cache_disabled(self, message):
        message["method"] = "Network.setResourceCachingDisabled"
        message["params"] = {"disabled": message["params"]["cacheDisabled"]}
        await self._send_message_to_target(message)

    async def _network_load_network_resource(self, message):
        await self.output_queue.put({"id": message["id"], "result": {"resource": {"success": True}}})

    async def _service_worker_enable(self, message):
        message["method"] = "Worker.enable"
        await self._send_message_to_target(message)

    async def _overlay_highlight_node(self, message):
        message["method"] = "DOM.highlightNode"
        await self._send_message_to_target(message)

    async def _runtime_compile_script(self, message):
        self._script_source_to_context_id[message["params"]["expression"]] = message["params"]["executionContextId"]
        response = await self.send_message_with_result(
            message["id"], "Runtime.parse", {"source": message["params"]["expression"]}
        )
        if response["result"]["result"] == "none":
            await self._simple_response(message, None)
            return
        lines = message["params"]["expression"][: response["result"]["range"]["endOffset"]].splitlines()
        lines = lines if lines else [""]
        await self.output_queue.put({
            "id": message["id"],
            "result": {
                "exceptionDetails": {
                    "exceptionId": 1,
                    "text": response["result"]["message"],
                    "lineNumber": len(lines) - 1,
                    "columnNumber": len(lines[-1]) - 1,
                }
            },
        })

    async def _runtime_get_isolate_id(self, message):
        await self.output_queue.put({"id": message["id"], "result": {"id": self._default_execution_id}})

    async def _target_set_auto_attach(self, message):
        await self._simple_response(message, None)
        await self.output_queue.put({
            "method": "Target.attachedToTarget",
            "params": {
                "sessionId": self.protocol.id_,
                "targetInfo": {"targetId": self.target_id},
                "waitingForDebugger": True,
            },
        })

    async def _css_take_computed_style_updates(self, message):
        await self.output_queue.put({"id": message["id"], "result": {"nodeIds": []}})

    async def _css_add_rule(self, message):
        message["params"]["selector"] = message["params"]["ruleText"].split("{")[0]
        await self._send_message_to_target(message)

    async def _input_emulate_touch_from_mouse_event(self, message):
        params = message["params"]
        if params["type"] == "mouseWheel":
            delta_x, delta_y = (
                params["deltaX"] // self.screencast.get_scale(),
                params["deltaY"] // self.screencast.get_scale(),
            )
            await self.evaluate_and_result(message["id"], f"window.scrollBy({-delta_x}, {-delta_y})")
        elif params["type"] == "mouseReleased":
            pass
        else:
            modifiers = params["modifiers"]
            x, y = params["x"] // self.screencast.get_scale(), params["y"] // self.screencast.get_scale()
            event_params = {
                "screenX": x,
                "screenY": y,
                "clientX": 0,
                "clientY": 0,
                "altKey": bool(modifiers & 1),
                "ctrlKey": bool(modifiers & 2),
                "metaKey": bool(modifiers & 4),
                "shiftKey": bool(modifiers & 8),
                "button": params["button"],
                "bubbles": True,
                "cancelable": False,
            }
            type_ = {"mousePressed": "click", "mouseReleased": "click", "mouseMoved": "mousemove"}[params["type"]]
            event_params = json.dumps(event_params)
            simulate_mouse_event = (
                "function simulateMouseEvent(type){"
                f"const element = document.elementFromPoint({x}, {y});"
                f"const e = new MouseEvent(type, JSON.parse('{event_params}'));"
                "element.dispatchEvent(e);"
                "element.focus();"
                "return e;}"
            )
            await self.evaluate_and_result(message["id"], f'({simulate_mouse_event})("{type_}")')
            if type_ == "click":
                await self.evaluate_and_result(message["id"], f'({simulate_mouse_event})("mouseup")')

        await self._simple_response(message, None)

    async def _input_dispatch_key_event(self, message):
        params = message["params"]
        key = params["key"]
        if params["type"] == "keyUp" and key == "Backspace":
            manipulation = "document.activeElement.value = document.activeElement.value.slice(0, -1);"
        elif params["type"] == "char" and key == "Enter":
            manipulation = (
                "var tagName = document.activeElement.tagName.toLowerCase();"
                'if (tagName === "textarea" || document.activeElement.isContentEditable) {'
                '    document.activeElement.value = document.activeElement.value + "\\n";'
                "} else {"
                '    const result = document.evaluate("./ancestor-or-self::form", document.activeElement, '
                "                                     null, XPathResult.FIRST_ORDERED_NODE_TYPE, null);"
                "    const e = result.singleNodeValue.ownerDocument.createEvent('Event');"
                '    e.initEvent("submit", true, true);'
                "    if (result.singleNodeValue.dispatchEvent(e)) { result.singleNodeValue.submit() }"
                "}"
            )
        elif params["type"] == "char":
            text = params["text"]
            manipulation = f'document.activeElement.value = document.activeElement.value + "{text}";'
        else:
            await self._simple_response(message, None)
            return

        simulate_key_event = (
            "function isEditable(element) {"
            "    if (element.disabled || element.readOnly)"
            "        return false;"
            "    var tagName = element.tagName.toLowerCase();"
            '    if (tagName === "textarea" || element.isContentEditable)'
            "        return true;"
            '    if (tagName != "input")'
            "        return false;"
            "    switch (element.type) {"
            '    case "color": case "date": case "datetime-local": case "email": case "file": case "month": '
            '    case "number": case "password": case "range": case "search": case "tel": case "text": case "time": '
            '    case "url": case "week":'
            "        return true;"
            "    }"
            "    return false;"
            "}"
            "if (isEditable(document.activeElement)) {"
            f"{manipulation}"
            "}"
        )
        await self.evaluate_and_result(message["id"], simulate_key_event)
        await self._simple_response(message, None)

    async def _target_created(self, message):
        self.target_id = message["params"]["targetInfo"]["targetId"]
        message["method"] = "Target.targetInfoChanged"
        message["params"]["targetInfo"]["url"] = await self.evaluate_and_result(1, "window.location.href")
        message["params"]["targetInfo"]["title"] = await self.evaluate_and_result(1, "document.title")
        message["params"]["targetInfo"]["attached"] = message["params"]["targetInfo"].pop("isProvisional")
        await self.output_queue.put(message)

    async def _target_destroyed(self, message):
        result = await self.send_message_with_result(1, "Page.getResourceTree", {})
        self.frame = result["result"]["frameTree"]["frame"]
        await self.output_queue.put({
            "method": "Page.frameNavigated",
            "params": {
                "frame": self.frame,
            },
        })
        await self.output_queue.put({
            "method": "Page.loadEventFired",
            "params": {
                "timestamp": datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%f"),
            },
        })
        await self.output_queue.put({"method": "DOM.documentUpdated"})

    async def _target_dispatch_message_from_target(self, message):
        message = json.loads(message["params"]["message"])
        if "error" in message:
            logger.error(message)
        if message.get("method", "") in self.to_cdp_special_dispatched_messages_methods:
            await self.to_cdp_special_dispatched_messages_methods[message["method"]](message)
        else:
            if "method" in message:
                logger.debug("DISPACHING", message["method"])
            await self.output_queue.put(message)

    async def _target_did_commit_provisional_target(self, message):
        pass

    async def _debugger_script_parsed(self, message):
        if not self._waiting_for_id:
            await self.output_queue.put(message)

    async def _debugger_script_failed_to_parse(self, message):
        message["params"] = {
            "endColumn": 0,
            "endLine": message["params"]["errorLine"],
            "executionContextId": self._script_source_to_context_id[message["params"]["scriptSource"]],
            "startColumn": 0,
            "startLine": message["params"]["startLine"],
            "url": message["params"]["url"],
            "scriptId": self._script_source_to_context_id[message["params"]["scriptSource"]],
            "hash": hashlib.sha1(message["params"]["scriptSource"]).hexdigest(),
        }
        await self.output_queue.put(message)

    async def _debugger_paused(self, message):
        message["params"]["reason"] = DEBUGGER_PAUSED_REASON[message["params"]["reason"]]
        if "breakpointId" in message["params"].get("data", {}):
            message["params"]["hitBreakpoints"] = [message["params"]["data"]["breakpointId"]]
        await self.output_queue.put(message)

    async def _debugger_global_object_cleared(self, message):
        await self.output_queue.put({"method": "DOM.documentUpdated"})

    async def _page_default_appearance_did_change(self, message):
        pass

    async def _runtime_execution_context_created(self, message):
        if message["params"]["context"]["type"] == "normal":
            self._default_execution_id = message["params"]["context"]["id"]
        message["params"] = {
            "context": {
                "id": message["params"]["context"]["id"],
                "origin": "default",
                "name": "",
                "uniqueId": message["params"]["context"]["frameId"],
            }
        }
        await self.output_queue.put(message)

    async def _console_message_added(self, message):
        log_record = {
            "source": LOG_MESSAGE_SOURCES[message["params"]["message"]["source"]],
            "level": LOG_MESSAGE_LEVELS[message["params"]["message"]["level"]],
            "text": message["params"]["message"]["text"],
            "timestamp": datetime.now().timestamp(),
        }
        if "url" in message["params"]["message"]:
            log_record["url"] = message["params"]["message"]["url"]
        if "line" in message["params"]["message"]:
            log_record["lineNumber"] = message["params"]["message"]["line"]
        if "networkRequestId" in message["params"]["message"]:
            log_record["networkRequestId"] = message["params"]["message"]["networkRequestId"]

        await self.output_queue.put({"method": "Log.entryAdded", "params": {"entry": log_record}})

    async def _network_response_received(self, message):
        params = message["params"]
        message["params"] = {
            "loaderId": params["loaderId"],
            "requestId": params["requestId"],
            "timestamp": params["timestamp"],
            "type": params["type"] if params["type"] in NETWORK_RESOURCE_TYPES else "Other",
            "response": {
                "url": params["response"]["url"],
                "status": params["response"]["status"],
                "statusText": params["response"]["statusText"],
                "headers": params["response"]["headers"],
                "mimeType": params["response"]["mimeType"],
                "connectionReused": False,
                "encodedDataLength": 0,
                "securityState": "unknown",
            },
        }
        if "frameId" in params:
            message["params"]["frameId"] = params["frameId"]
        await self.output_queue.put(message)

    async def _network_loading_finished(self, message):
        params = message["params"]
        header_size = params["metrics"].get("responseHeaderBytesReceived", 0)
        body_size = params["metrics"].get("responseBodyBytesReceived", 0)
        message["params"] = {
            "encodedDataLength": header_size + body_size,
            "requestId": params["requestId"],
            "timestamp": params["timestamp"],
        }
        await self.output_queue.put(message)



from pymobiledevice3.services.dvt.instruments import ChannelService
from pymobiledevice3.services.dvt.instruments.location_simulation_base import LocationSimulationBase
from pymobiledevice3.services.remote_server import MessageAux


class LocationSimulation(LocationSimulationBase, ChannelService):
    IDENTIFIER = "com.apple.instruments.server.services.LocationSimulation"

    def __init__(self, dvt):
        super().__init__()
        ChannelService.__init__(self, dvt)

    async def set(self, latitude: float, longitude: float) -> None:
        channel = await self._channel_ref()
        await channel.simulateLocationWithLatitude_longitude_(MessageAux().append_obj(latitude).append_obj(longitude))
        await channel.receive_plist()

    async def clear(self) -> None:
        channel = await self._channel_ref()
        await channel.stopLocationSimulation()



import dataclasses
import typing
from typing import Optional

from pymobiledevice3.exceptions import DisableMemoryLimitError
from pymobiledevice3.osu.os_utils import get_os_utils
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments import ChannelService
from pymobiledevice3.services.remote_server import MessageAux

OSUTIL = get_os_utils()


@dataclasses.dataclass
class OutputReceivedEvent:
    pid: int
    date: int
    message: str

    @classmethod
    def create(cls, message) -> "OutputReceivedEvent":
        try:
            date = OSUTIL.parse_timestamp(message[2].value)
        except (ValueError, OSError):
            date = None

        return cls(pid=message[1].value, date=date, message=message[0].value)


class ProcessControl(ChannelService):
    IDENTIFIER = "com.apple.instruments.server.services.processcontrol"

    def __init__(self, dvt: DvtSecureSocketProxyService):
        super().__init__(dvt)

    async def signal(self, pid: int, sig: int):
        """
        Send signal to process
        :param pid: PID of process to send signal.
        :param sig: SIGNAL to send
        """
        channel = await self._channel_ref()
        await channel.sendSignal_toPid_(MessageAux().append_obj(sig).append_obj(pid), expects_reply=True)
        return await channel.receive_plist()

    async def disable_memory_limit_for_pid(self, pid: int) -> None:
        """
        Waive memory limit for a given pid
        :param pid: process id.
        """
        channel = await self._channel_ref()
        await channel.requestDisableMemoryLimitsForPid_(MessageAux().append_int(pid), expects_reply=True)
        if not await channel.receive_plist():
            raise DisableMemoryLimitError()

    async def kill(self, pid: int):
        """
        Kill a process.
        :param pid: PID of process to kill.
        """
        channel = await self._channel_ref()
        await channel.killPid_(MessageAux().append_obj(pid), expects_reply=False)

    async def process_identifier_for_bundle_identifier(self, app_bundle_identifier: str) -> int:
        channel = await self._channel_ref()
        await channel.processIdentifierForBundleIdentifier_(
            MessageAux().append_obj(app_bundle_identifier), expects_reply=True
        )
        return await channel.receive_plist()

    async def launch(
        self,
        bundle_id: str,
        arguments=None,
        kill_existing: bool = True,
        start_suspended: bool = False,
        environment: Optional[dict] = None,
        extra_options: Optional[dict] = None,
    ) -> int:
        """
        Launch a process.
        :param bundle_id: Bundle id of the process.
        :param list arguments: List of argument to pass to process.
        :param kill_existing: Whether to kill an existing instance of this process.
        :param start_suspended: Same as WaitForDebugger.
        :param environment: Environment variables to pass to process.
        :param extra_options: Extra options to pass to process.
        :return: PID of created process.
        """
        arguments = [] if arguments is None else arguments
        environment = {} if environment is None else environment
        options = {
            "StartSuspendedKey": start_suspended,
            "KillExisting": kill_existing,
        }
        if extra_options:
            options.update(extra_options)
        args = (
            MessageAux()
            .append_obj("")
            .append_obj(bundle_id)
            .append_obj(environment)
            .append_obj(arguments)
            .append_obj(options)
        )
        channel = await self._channel_ref()
        await channel.launchSuspendedProcessWithDevicePath_bundleIdentifier_environment_arguments_options_(args)
        result = await channel.receive_plist()
        assert result
        return result

    async def __aiter__(self) -> typing.AsyncGenerator[OutputReceivedEvent, None]:
        while True:
            key, value = await self._receive_key_value()
            if key == "outputReceived:fromProcess:atTime:":
                yield OutputReceivedEvent.create(value)

    async def _receive_key_value(self):
        channel = await self._channel_ref()
        return await channel.receive_key_value()



import logging
import os
import site
import sys
from pathlib import Path

import coloredlogs
import PyInstaller.__main__

import pymobiledevice3.__main__
import pymobiledevice3.cli
import pymobiledevice3.resources

coloredlogs.install(level=logging.DEBUG)

ROOT = Path(__file__).parent.parent.parent
DEFAULT_OUTPUT = ROOT / Path("dist/__main__").with_suffix(".exe" if sys.platform == "win32" else "")
OUTPUT = (ROOT / "dist/pymobiledevice3").with_suffix(".exe" if sys.platform == "win32" else "")


def main() -> None:
    site_packages_path = site.getsitepackages()[0]
    resources_dir = Path(pymobiledevice3.resources.__file__).parent
    pymobiledevice3_cli_path = Path(pymobiledevice3.cli.__file__).parent

    hidden_imports = []

    for module in pymobiledevice3_cli_path.iterdir():
        if module.name.endswith(".py") and module.name != "__init__.py":  # Avoid including the __init__.py
            # Create the module name to be added to hidden imports
            module_name = "pymobiledevice3.cli." + os.path.splitext(module.name)[0]
            hidden_imports.append("--hidden-import=" + module_name)

    pyinstaller_args = [
        pymobiledevice3.__main__.__file__,
        "--hidden-import=ipsw_parser",
        "--hidden-import=zeroconf",
        "--hidden-import=pyimg4",
        "--hidden-import=zeroconf._utils.ipaddress",
        "--hidden-import=zeroconf._handlers.answers",
        "--hidden-import=readchar",
        "--copy-metadata=pyimg4",
        "--copy-metadata=readchar",
        "--copy-metadata=apple_compress",
        "--onefile",
    ]
    if sys.platform == "win32":
        pyinstaller_args.extend([
            "--add-binary",
            f"{site_packages_path}/Lib/site-packages/pytun_pmd3/wintun/*;pytun_pmd3/wintun/bin",
            "--add-binary",
            f"{resources_dir}/webinspector;pymobiledevice3/resources/webinspector",
        ])
    else:
        pyinstaller_args.extend([
            "--add-binary",
            f"{site_packages_path}/pytun_pmd3:pytun_pmd3",
            "--add-binary",
            f"{resources_dir}/webinspector:pymobiledevice3/resources/webinspector",
        ])
    pyinstaller_args.extend(hidden_imports)
    PyInstaller.__main__.run(pyinstaller_args)
    DEFAULT_OUTPUT.rename(OUTPUT)


if __name__ == "__main__":
    main()



import logging

from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command, print_json
from pymobiledevice3.services.amfi import AmfiService

logger = logging.getLogger(__name__)


cli = InjectingTyper(
    name="amfi",
    help="Enable developer-mode or query its state",
    no_args_is_help=True,
)


@cli.command()
@async_command
async def reveal_developer_mode(service_provider: ServiceProviderDep) -> None:
    """reveal developer mode option in device's UI"""
    await AmfiService(service_provider).reveal_developer_mode_option_in_ui()


@cli.command()
@async_command
async def enable_developer_mode(service_provider: ServiceProviderDep) -> None:
    """enable developer mode"""
    if await service_provider.get_developer_mode_status():
        logger.info("Developer mode is already enabled")
        return
    await AmfiService(service_provider).enable_developer_mode()


@cli.command()
@async_command
async def developer_mode_status(service_provider: ServiceProviderDep) -> None:
    """query developer mode status"""
    print_json(await service_provider.get_developer_mode_status())



import asyncio
import logging
import tempfile
from typing import Annotated, Optional

import typer
from typer_injector import InjectingTyper

from pymobiledevice3 import usbmux
from pymobiledevice3.cli.cli_common import USBMUX_ENV_VAR, USBMUX_OPTION_HELP, async_command, print_json
from pymobiledevice3.lockdown import create_using_usbmux
from pymobiledevice3.tcp_forwarder import UsbmuxTcpForwarder

logger = logging.getLogger(__name__)


cli = InjectingTyper(
    name="usbmux",
    help="Inspect usbmuxd-connected devices and forward TCP ports to them.",
    no_args_is_help=True,
)


@cli.command("forward")
@async_command
async def usbmux_forward(
    src_port: Annotated[
        int,
        typer.Argument(min=1, max=0xFFFF),
    ],
    dst_port: Annotated[
        int,
        typer.Argument(min=1, max=0xFFFF),
    ],
    *,
    usbmux_address: Annotated[
        Optional[str],
        typer.Option(
            "--usbmux",
            envvar=USBMUX_ENV_VAR,
            help=USBMUX_OPTION_HELP,
        ),
    ] = None,
    serial: Annotated[
        Optional[str],
        typer.Option("--serial", "--udid", help="Device serial/UDID to forward traffic to."),
    ] = None,
    host: Annotated[
        str,
        typer.Option(help="Address to bind the local port to."),
    ] = "127.0.0.1",
    daemonize: Annotated[
        bool,
        typer.Option("--daemonize", "-d", help="Run the forwarder in the background."),
    ] = False,
) -> None:
    """Forward a local TCP port to the device via usbmuxd."""
    forwarder = UsbmuxTcpForwarder(serial, dst_port, src_port, usbmux_address=usbmux_address)

    if daemonize:
        try:
            from daemonize import Daemonize
        except ImportError as e:
            raise NotImplementedError("daemonizing is only supported on unix platforms") from e

        with tempfile.NamedTemporaryFile("wt") as pid_file:
            daemon = Daemonize(
                app=f"forwarder {src_port}->{dst_port}",
                pid=pid_file.name,
                action=lambda: asyncio.run(forwarder.start(address=host)),
            )
            daemon.start()
    else:
        await forwarder.start(address=host)


@cli.command("list")
@async_command
async def usbmux_list(
    usbmux_address: Annotated[
        Optional[str],
        typer.Option(
            "--usbmux",
            envvar=USBMUX_ENV_VAR,
            help=USBMUX_OPTION_HELP,
        ),
    ] = None,
    usb: Annotated[
        bool,
        typer.Option(
            "--usb",
            "-u",
            help="show only USB devices",
        ),
    ] = False,
    network: Annotated[
        bool,
        typer.Option(
            "--network",
            "-n",
            help="show only network devices",
        ),
    ] = False,
    simple: Annotated[
        bool,
        typer.Option(
            "--simple",
            help="List only UDIDs without connecting to lockdownd.",
        ),
    ] = False,
) -> None:
    """List devices known to usbmuxd (USB and Wi-Fi)."""
    connected_devices = []
    for device in await usbmux.list_devices(usbmux_address=usbmux_address):
        udid = device.serial

        if usb and not device.is_usb:
            continue

        if network and not device.is_network:
            continue

        if simple:
            connected_devices.append(udid)
            continue

        lockdown = await create_using_usbmux(
            udid, autopair=False, connection_type=device.connection_type, usbmux_address=usbmux_address
        )
        connected_devices.append(lockdown.short_info)

    print_json(connected_devices)



import asyncio
import logging
import threading
from typing import Any, Optional

from bpylist2 import archiver
from packaging.version import Version

from pymobiledevice3.exceptions import AppNotInstalledError, ConnectionTerminatedError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.dvt_testmanaged_proxy import DvtTestmanagedProxyService
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl
from pymobiledevice3.services.house_arrest import HouseArrestService
from pymobiledevice3.services.installation_proxy import InstallationProxyService
from pymobiledevice3.services.remote_server import (
    NSURL,
    NSUUID,
    Channel,
    MessageAux,
    RemoteServer,
    XCTCapabilities,
    XCTestConfiguration,
)

logger = logging.getLogger(__name__)


class ReverseRemoteService:
    def __init__(self, service: RemoteServer, channel: int):
        self.service = service
        self.channel = channel
        self.logger = logger.getChild(self.__class__.__name__)

    async def __on_method_call__(
        self, selector: str, args: Optional[MessageAux] = None, message_id: Optional[int] = None
    ):
        if message_id is None:
            raise ValueError("message_id is required for reverse DTX replies")
        _selector = selector.replace(":", "_")
        if not hasattr(self, _selector):
            logger.debug(
                "%s: unhandled method call: selector='%s', _selector='%s', args='%s'",
                self.__class__.__name__,
                selector,
                _selector,
                args,
            )
            await self.service.send_reply_error(self.channel, message_id)
            return
        ret = getattr(self, _selector)(args)
        await self.service.send_reply(self.channel, message_id, ret)


class XCUITestListener(ReverseRemoteService):
    """dtxproxy:XCTestDriverInterface:XCTestManager_IDEInterface"""

    def __init__(
        self,
        service: DvtTestmanagedProxyService,
        channel: int,
        test_configuration: XCTestConfiguration,
    ):
        super().__init__(service, channel)
        self.test_config = test_configuration

    def __on_method_call__(self, selector: str, args: Optional[MessageAux] = None, message_id: Optional[int] = None):
        return super().__on_method_call__(selector.removeprefix("_XCT_"), args, message_id)

    def logDebugMessage_(self, args: Optional[MessageAux] = None):
        self.logger.debug("logDebugMessage: %s", args[0].value.strip())

    def testRunnerReadyWithCapabilities_(self, args: Optional[MessageAux] = None):
        self.logger.debug("testRunnerReadyWithCapabilities: %s", args[0].value)
        return self.test_config

    def didFinishExecutingTestPlan(self, args: Optional[MessageAux] = None):
        self.logger.debug("didFinishExecutingTestPlan")


class XCUITestPlanConsumer:
    def __init__(
        self,
        pid: int,
        pctrl: ProcessControl,
        ctrl_dvt: DvtTestmanagedProxyService,
        ctrl_chan: Channel,
        main_dvt: DvtTestmanagedProxyService,
        main_chan: Channel,
        test_configuration: XCTestConfiguration,
    ):
        self.pid = pid
        self.pctrl = pctrl
        self.ctrl_dvt = ctrl_dvt
        self.ctrl_chan = ctrl_chan
        self.main_dvt = main_dvt
        self.main_chan = main_chan
        self.xctest_config = test_configuration
        self.listener: Optional[XCUITestListener] = None
        self.__running__ = False
        self.__closing__ = False
        self.completed = threading.Event()

    def set_listener(self, listener: XCUITestListener):
        self.listener = listener

    async def consume(self):
        if self.listener is None:
            self.listener = XCUITestListener(self.main_dvt, self.main_chan, self.xctest_config)
        try:
            self.__running__ = True
            while self.__running__:
                (key, value), mheader = await self.main_dvt.recv_plist(self.main_chan, return_header=True)
                if key == "_XCT_didFinishExecutingTestPlan":
                    self.__closing__ = True
                await self.listener.__on_method_call__(key, value, mheader.identifier)
        except (ConnectionTerminatedError, ConnectionResetError):
            if not self.__closing__:
                logger.warning("connection aborted")
            return
        except (InterruptedError, KeyboardInterrupt):
            return
        finally:
            self.completed.set()

    async def stop(self):
        if not self.__running__:
            return

        self.__closing__ = True
        self.__running__ = False

        logger.debug("Killing UITest with pid %d ...", self.pid)
        try:
            await self.pctrl.kill(self.pid)
        except RuntimeError as e:
            if "no running event loop" not in str(e):
                raise
            logger.debug("Skipping process kill during event-loop teardown")

        for dvt in (self.main_dvt, self.ctrl_dvt):
            try:
                await dvt.close()
            except RuntimeError as e:
                if "no running event loop" not in str(e):
                    raise
                logger.debug("Skipping DVT close during event-loop teardown")


class XCUITestService:
    IDENTIFIER = "dtxproxy:XCTestManager_IDEInterface:XCTestManager_DaemonConnectionInterface"
    XCODE_VERSION = 36  # not important
    __TESTDRIVERINTERFACE_CHANNEL__ = -1  # a special channel that is reserved for "XCTestDriverInterface"

    def __init__(self, service_provider: LockdownServiceProvider):
        self.service_provider = service_provider
        self.pctl: Optional[ProcessControl] = None
        self.product_major_version = Version(service_provider.product_version).major

    async def start(
        self,
        bundle_id: str,
        test_runner_env: Optional[dict] = None,
        test_runner_args: Optional[list] = None,
    ) -> XCUITestPlanConsumer:
        if self.pctl is None:
            self.pctl = await self.init_process_control()
        session_identifier = NSUUID.uuid4()
        app_info = await get_app_info(self.service_provider, bundle_id)

        xctest_configuration = generate_xctestconfiguration(
            app_info, session_identifier, self.product_major_version, test_runner_env, test_runner_args
        )
        xctest_path = f"/tmp/{str(session_identifier).upper()}.xctestconfiguration"  # yapf: disable

        if self.product_major_version < 17:
            await self.setup_xcuitest(bundle_id, xctest_path, xctest_configuration)

        ctrl_dvt, ctrl_chan, main_dvt, main_chan = await self.init_ide_channels(
            session_identifier, xctest_configuration._config["IDECapabilities"]
        )

        pid = await self.launch_test_app(
            session_identifier, app_info, bundle_id, xctest_path, test_runner_env, test_runner_args
        )
        logger.debug("Runner started with pid:%d, waiting for testBundleReady", pid)

        if self.product_major_version < 17:
            await asyncio.sleep(1)

        await self.authorize_test_process_id(ctrl_chan, pid)
        # acknoledge that we will have this service listening on our channel
        await main_dvt.serve_channel("dtxproxy:XCTestDriverInterface:XCTestManager_IDEInterface", main_chan)
        await self.start_executing_test_plan_with_protocol_version(main_dvt, self.XCODE_VERSION)

        # TODO: broadcast channel side messages are not consumed here.
        return XCUITestPlanConsumer(pid, self.pctl, ctrl_dvt, ctrl_chan, main_dvt, main_chan, xctest_configuration)

    async def run(
        self,
        bundle_id: str,
        test_runner_env: Optional[dict] = None,
        test_runner_args: Optional[list] = None,
    ):
        consumer = await self.start(bundle_id, test_runner_env, test_runner_args)
        try:
            await consumer.consume()
        finally:
            await consumer.stop()

    async def init_process_control(self):
        dvt_proxy = DvtSecureSocketProxyService(lockdown=self.service_provider)
        await dvt_proxy.perform_handshake()
        return ProcessControl(dvt_proxy)

    async def init_ide_channels(self, session_identifier: NSUUID, ide_capabilities: XCTCapabilities):
        """
        Initialize the two channels required by XcodeIDE to control the test session. The first channel is used to authorize the test process, and the second channel is used to manage the test.

        return: (ctrl_dvt, ctrl_chan, main_dvt, main_chan)
         - ctrl_dvt: the DvtTestmanagedProxyService instance for the control channel
         - ctrl_chan: the Channel instance for the control channel
         - main_dvt: the DvtTestmanagedProxyService instance for the main channel
         - main_chan: the Channel instance for the main channel
        """
        # XcodeIDE require two connections
        ctrl_dvt = DvtTestmanagedProxyService(lockdown=self.service_provider)
        await ctrl_dvt.perform_handshake()

        logger.debug("make channel %s", self.IDENTIFIER)
        ctrl_chan = await ctrl_dvt.make_channel(self.IDENTIFIER)
        if self.product_major_version >= 17:
            await ctrl_dvt.send_message(
                ctrl_chan,
                "_IDE_initiateControlSessionWithCapabilities:",
                MessageAux().append_obj(XCTCapabilities({})),
            )
            reply = await ctrl_chan.receive_plist()
            logger.debug("ctrl_conn handshake capabilities: %s", reply)
        elif self.product_major_version >= 11:
            await ctrl_dvt.send_message(
                ctrl_chan,
                "_IDE_initiateControlSessionWithProtocolVersion:",
                MessageAux().append_obj(self.XCODE_VERSION),
            )
            reply = await ctrl_chan.receive_plist()
            logger.debug("ctrl_conn handshake xcode version: %s", reply)

        main_dvt = DvtTestmanagedProxyService(lockdown=self.service_provider)
        await main_dvt.perform_handshake()
        main_chan = await main_dvt.make_channel(self.IDENTIFIER)
        if self.product_major_version >= 17:
            await main_dvt.send_message(
                channel=main_chan,
                selector="_IDE_initiateSessionWithIdentifier:capabilities:",
                args=MessageAux().append_obj(session_identifier).append_obj(ide_capabilities),
            )
            reply = await main_chan.receive_plist()
            logger.debug("main_conn handshake capabilities: %s", reply)
        else:
            await main_dvt.send_message(
                channel=main_chan,
                selector="_IDE_initiateSessionWithIdentifier:forClient:atPath:protocolVersion:",
                args=MessageAux()
                .append_obj(session_identifier)
                .append_obj("not-very-import-part")  # this part is not important
                .append_obj("/Applications/Xcode.app/Contents/Developer/usr/bin/xcodebuild")
                .append_obj(self.XCODE_VERSION),
            )
            reply = await main_chan.receive_plist()
            logger.debug("main_conn handshake xcode version: %s", reply)
        return ctrl_dvt, ctrl_chan, main_dvt, main_chan

    async def setup_xcuitest(
        self,
        bundle_id: str,
        xctest_path: str,
        xctest_configuration: XCTestConfiguration,
    ):
        """push xctestconfiguration to app VendDocuments"""
        async with await HouseArrestService.create(lockdown=self.service_provider, bundle_id=bundle_id) as afc:
            for name in await afc.listdir("/tmp"):
                if name.endswith(".xctestconfiguration"):
                    logger.debug("remove /tmp/%s", name)
                    await afc.rm("/tmp/" + name)
            await afc.set_file_contents(xctest_path, archiver.archive(xctest_configuration))

    async def start_executing_test_plan_with_protocol_version(
        self, dvt: DvtTestmanagedProxyService, protocol_version: int
    ):
        await dvt.send_message(
            self.__TESTDRIVERINTERFACE_CHANNEL__,
            "_IDE_startExecutingTestPlanWithProtocolVersion:",
            MessageAux().append_obj(protocol_version),
            expects_reply=False,
        )

    async def authorize_test_process_id(self, chan: Channel, pid: int):
        selector = None
        aux = MessageAux()
        if self.product_major_version >= 12:
            selector = "_IDE_authorizeTestSessionWithProcessID:"
            aux.append_obj(pid)
        elif self.product_major_version >= 10:
            selector = "_IDE_initiateControlSessionForTestProcessID:protocolVersion:"
            aux.append_obj(pid)
            aux.append_obj(self.XCODE_VERSION)
        else:
            selector = "_IDE_initiateControlSessionForTestProcessID:"
            aux.append_obj(pid)
        await chan.send_message(selector, aux)
        reply = await chan.receive_plist()
        if isinstance(reply, bool) and reply is True:
            logger.debug("authorizing test session for pid %d successful %r", pid, reply)
        else:
            raise RuntimeError(f"Failed to authorize test process id: {reply}")

    async def launch_test_app(
        self,
        test_session_identifier: NSUUID,
        app_info: dict,
        bundle_id: str,
        xctest_path: str,
        test_runner_env: Optional[dict] = None,
        test_runner_args: Optional[list] = None,
    ) -> int:
        app_container = app_info["Container"]
        app_path = app_info["Path"]
        exec_name = app_info["CFBundleExecutable"]
        # # logger.debug()('CFBundleExecutable: %s', exec_name)
        # # CFBundleName always endswith -Runner
        assert exec_name.endswith("-Runner"), f"Invalid CFBundleExecutable: {exec_name}"
        target_name = exec_name[: -len("-Runner")]

        app_env = {
            "CA_ASSERT_MAIN_THREAD_TRANSACTIONS": "0",
            "CA_DEBUG_TRANSACTIONS": "0",
            "DYLD_FRAMEWORK_PATH": app_path + "/Frameworks:",
            "DYLD_LIBRARY_PATH": app_path + "/Frameworks",
            "MTC_CRASH_ON_REPORT": "1",
            "NSUnbufferedIO": "YES",
            "SQLITE_ENABLE_THREAD_ASSERTIONS": "1",
            "WDA_PRODUCT_BUNDLE_IDENTIFIER": "",
            "XCTestBundlePath": f"{app_info['Path']}/PlugIns/{target_name}.xctest",
            "XCTestConfigurationFilePath": app_container + xctest_path,
            "XCODE_DBG_XPC_EXCLUSIONS": "com.apple.dt.xctestSymbolicator",
            # the following maybe no needed
            # 'MJPEG_SERVER_PORT': '',
            # 'USE_PORT': '',
            # 'LLVM_PROFILE_FILE': app_container + '/tmp/%p.profraw', # %p means pid
            "XCTestSessionIdentifier": str(test_session_identifier).upper(),
        }

        if self.product_major_version >= 11:
            app_env["DYLD_INSERT_LIBRARIES"] = "/Developer/usr/lib/libMainThreadChecker.dylib"
            app_env["OS_ACTIVITY_DT_MODE"] = "YES"
        if self.product_major_version >= 17:
            app_env["DYLD_FRAMEWORK_PATH"] = f"${app_env['DYLD_FRAMEWORK_PATH']}/System/Developer/Library/Frameworks:"
            app_env["DYLD_LIBRARY_PATH"] = f"${app_env['DYLD_LIBRARY_PATH']}:/System/Developer/usr/lib"
            app_env["XCTestConfigurationFilePath"] = ""  # sent as return value of _XCT_testRunnerReadyWithCapabilities
            app_env["XCTestManagerVariant"] = "DDI"

        if test_runner_env:
            app_env.update(test_runner_env)

        app_args = [
            "-NSTreatUnknownArgumentsAsOpen",
            "NO",
            "-ApplePersistenceIgnoreState",
            "YES",
        ]
        app_args.extend(test_runner_args or [])
        app_options = {"StartSuspendedKey": False}
        if self.product_major_version >= 12:
            app_options["ActivateSuspended"] = True

        # go-ios uses the following to spawn a new process for the test runner.
        # async def launch_app():
        #     async with AppServiceService(self.service_provider) as app_service:
        #         resp = await app_service.launch_application(bundle_id, arguments=app_args, environment=app_env, extra_options=app_options, kill_existing=True)
        #         return int(resp["processToken"]["processIdentifier"])
        # pid = await launch_app()
        return await self.pctl.launch(
            bundle_id,
            arguments=app_args,
            environment=app_env,
            extra_options=app_options,
        )


async def start_xcuitest_runner(
    service_provider: LockdownServiceProvider,
    bundle_id: str,
    test_runner_env: Optional[dict] = None,
    test_runner_args: Optional[list] = None,
) -> tuple[XCUITestPlanConsumer, asyncio.Task]:
    service = XCUITestService(service_provider)
    consumer = await service.start(bundle_id, test_runner_env, test_runner_args)
    task = asyncio.create_task(consumer.consume())
    return consumer, task


async def get_app_info(service_provider: LockdownServiceProvider, bundle_id: str) -> dict[str, Any]:
    async with InstallationProxyService(lockdown=service_provider) as install_service:
        apps = await install_service.get_apps(bundle_identifiers=[bundle_id])
        if not apps:
            raise AppNotInstalledError(f"No app with bundle id {bundle_id} found")
        return apps[bundle_id]


def generate_xctestconfiguration(
    target_app_info: dict,
    session_identifier: NSUUID,
    product_major_version: int,
    target_app_env: Optional[dict] = None,
    target_app_args: Optional[list] = None,
    tests_to_run: Optional[list] = None,
    tests_to_skip: Optional[list] = None,
) -> XCTestConfiguration:
    exec_name: str = target_app_info["CFBundleExecutable"]
    assert exec_name.endswith("-Runner"), f"Invalid CFBundleExecutable: {exec_name}"
    config_name = exec_name[: -len("-Runner")]

    automation_framework_path = "/Developer/Library/PrivateFrameworks/XCTAutomationSupport.framework"
    if product_major_version >= 17:
        automation_framework_path = "/System/Developer/Library/PrivateFrameworks/XCTAutomationSupport.framework"

    return XCTestConfiguration({
        "testBundleURL": NSURL(None, f"file://{target_app_info['Path']}/PlugIns/{config_name}.xctest"),
        "sessionIdentifier": session_identifier,
        "productModuleName": config_name,
        "targetApplicationBundleID": target_app_info.get("CFBundleIdentifier"),
        "targetApplicationPath": target_app_info.get("Path"),
        "targetApplicationEnvironment": target_app_env or {},
        "targetApplicationArguments": target_app_args or [],
        "testsToRun": tests_to_run or set(),
        "testsToSkip": tests_to_skip or set(),
        "testsMustRunOnMainThread": True,
        "reportResultsToIDE": True,
        "reportActivities": True,
        "automationFrameworkPath": automation_framework_path,
        "testApplicationDependencies": None,
    })



from pymobiledevice3.services.dvt.instruments import ChannelService
from pymobiledevice3.services.remote_server import MessageAux


class ApplicationListing(ChannelService):
    IDENTIFIER = "com.apple.instruments.server.services.device.applictionListing"

    def __init__(self, dvt):
        super().__init__(dvt)

    async def applist(self) -> list:
        """
        Get the applications list from the device.
        :return: List of applications and their attributes.
        """
        channel = await self._channel_ref()
        await channel.installedApplicationsMatching_registerUpdateToken_(MessageAux().append_obj({}).append_obj(""))
        result = await channel.receive_plist()
        assert isinstance(result, list)
        return result



import json
import logging
import os.path
from uuid import UUID

import click
import coloredlogs

MAGIC = b"\x0b\x10\x00\x00"
DYLD_MAGIC = b"dyld_v1"
MAP_FILENAME = os.path.join(os.path.dirname(__file__), "dsc_uuid_map.json")
PARTITIONS = ("/System", "/usr", "/Applications", "/private")
DYLD_UUID_OFFSET = 0x58
UUID_SIZE = 0x10

logger = logging.getLogger(__name__)


def get_dsc_map(dsc_uuid):
    with open(MAP_FILENAME) as f:
        uuid_map = json.load(f)

    return uuid_map.get(dsc_uuid)


def sanitize_path(path):
    for partition in PARTITIONS:
        if path.startswith(partition):
            return path

    for partition in PARTITIONS:
        if partition in path:
            return partition + path.split(partition, 1)[1]


@click.command()
@click.argument("dyld_uuid", type=click.UUID)
@click.argument("dsc", type=click.File("rb"))
@click.option("-f", "--force", is_flag=True)
def main(dsc, dyld_uuid, force):
    """
    Simple utility to get all UUIDs used for symbolication from given DSC.
    The UUID of `/usr/lib/dyld` still needs manual insertion.
    """
    with open(MAP_FILENAME) as f:
        uuid_map = json.load(f)

    dsc = dsc.read()

    if not dsc.startswith(DYLD_MAGIC):
        logging.error("invalid dsc file")
        return

    dsc_uuid = str(UUID(bytes=dsc[DYLD_UUID_OFFSET : DYLD_UUID_OFFSET + UUID_SIZE]))

    if dsc_uuid in uuid_map:
        logger.warning(f"dsc {dsc_uuid} is already found in dsc_uuid_map")
        if not force:
            logger.info("exiting. use --force to force update")
            return
    else:
        uuid_map[dsc_uuid] = {str(dyld_uuid): "/usr/lib/dyld"}

    for i in range(0, len(dsc) - 4, 4):
        # we can assume MAGIC is always aligned to 4

        if dsc[i : i + 4] != MAGIC:
            continue

        # skip NULLs for filename pad
        j = i - 1
        while dsc[j] == 0:
            j -= 1

        # read filename backwards
        filename = ""
        c = chr(dsc[j])
        while c.isprintable():
            filename = c + filename
            j -= 1
            c = chr(dsc[j])

        if "/" not in filename:
            continue

        filename = sanitize_path(filename)

        # read uuid
        uuid = UUID(bytes=dsc[i + 4 : i + 4 + UUID_SIZE])

        logging.info(f"offset: 0x{i:x} image: {filename} uuid: {uuid}")

        uuid_map[dsc_uuid][str(uuid)] = filename

    with open(MAP_FILENAME, "w") as f:
        json.dump(uuid_map, f, indent=4)


if __name__ == "__main__":
    coloredlogs.install(level=logging.DEBUG)
    main()



import pytest
import pytest_asyncio

from pymobiledevice3.exceptions import AppNotInstalledError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.house_arrest import HouseArrestService
from pymobiledevice3.services.installation_proxy import InstallationProxyService


@pytest_asyncio.fixture(scope="function")
async def user_bundle_id(lockdown: LockdownClient) -> str:
    async with InstallationProxyService(lockdown=lockdown) as installation_proxy:
        user_apps = await installation_proxy.get_apps(application_type="User")
    if not user_apps:
        pytest.skip("No user apps installed to exercise house arrest")
    return next(iter(user_apps))


@pytest.mark.asyncio
async def test_missing_bundle_id(lockdown: LockdownClient) -> None:
    with pytest.raises(AppNotInstalledError):
        async with await HouseArrestService.create(lockdown=lockdown, bundle_id="com.pymobiledevice3.missing.app"):
            pass


@pytest.mark.asyncio
async def test_vend_container_lists_app_root(lockdown: LockdownClient, user_bundle_id: str) -> None:
    async with await HouseArrestService.create(
        lockdown=lockdown, bundle_id=user_bundle_id, documents_only=False
    ) as service:
        entries = set(await service.listdir("/"))
    assert "Documents" in entries



from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments import ChannelService


class Screenshot(ChannelService):
    IDENTIFIER = "com.apple.instruments.server.services.screenshot"

    def __init__(self, dvt: DvtSecureSocketProxyService):
        super().__init__(dvt)

    async def get_screenshot(self) -> bytes:
        """get device screenshot"""
        channel = await self._channel_ref()
        await channel.takeScreenshot(expects_reply=True)
        return await channel.receive_plist()



from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import OSUTILS, ServiceProviderDep, async_command, print_json
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.condition_inducer import ConditionInducer

cli = InjectingTyper(
    name="developer",
    help="Force predefined device conditions (network, thermal, battery) via DVT.",
    no_args_is_help=True,
)


@cli.command("list")
@async_command
async def condition_list(service_provider: ServiceProviderDep) -> None:
    """List available condition profiles."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        print_json(await ConditionInducer(dvt).list())


@cli.command("clear")
@async_command
async def condition_clear(service_provider: ServiceProviderDep) -> None:
    """Clear any active induced condition."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        await ConditionInducer(dvt).clear()


@cli.command("set")
@async_command
async def condition_set(service_provider: ServiceProviderDep, profile_identifier: str) -> None:
    """Apply a specific condition profile by identifier."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        await ConditionInducer(dvt).set(profile_identifier)
        OSUTILS.wait_return()



from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.screenshot import Screenshot

PNG_HEADER = b"\x89\x50\x4e\x47\x0d\x0a\x1a\x0a"
TIFF_HEADER = b"\x4d\x4d\x00\x2a"


async def test_screenshot(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test that taking a screenshot returns a PNG.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    screenshot = await Screenshot(dvt).get_screenshot()
    assert screenshot.startswith(PNG_HEADER) or screenshot.startswith(TIFF_HEADER)



import asyncio
import copy
import io
import logging
import os
import plistlib
import socket
import uuid
from collections.abc import Awaitable
from contextlib import suppress
from functools import partial
from pprint import pprint
from typing import Any, Callable, ClassVar, Optional, Union

from bpylist2 import archiver
from construct import (
    Adapter,
    Const,
    Container,
    Default,
    GreedyBytes,
    GreedyRange,
    Int16ul,
    Int32sl,
    Int32ul,
    Int64ul,
    Prefixed,
    Select,
    Struct,
    Switch,
    this,
)
from pygments import formatters, highlight, lexers

from pymobiledevice3.exceptions import (
    ChannelClosedError,
    ConnectionTerminatedError,
    DvtException,
    UnrecognizedSelectorError,
)
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService
from pymobiledevice3.utils import start_ipython_shell

SHELL_USAGE = """
# This shell allows you to send messages to the DVTSecureSocketProxy and receive answers easily.
# Generally speaking, each channel represents a group of actions.
# Calling actions is done using a selector and auxiliary (parameters).
# Receiving answers is done by getting a return value and seldom auxiliary (private / extra parameters).
# To see the available channels, type the following:
developer.supported_identifiers

# In order to send messages, you need to create a channel:
channel = await developer.make_channel('com.apple.instruments.server.services.deviceinfo')

# After creating the channel you can call allowed selectors:
await channel.runningProcesses()

# If an answer is expected, you can receive it using the receive method:
processes = await channel.receive_plist()

# You can also call methods from the broadcast channel
await broadcast.someMethod()

# Sometimes the selector requires parameters, You can add them using MessageAux. For example lets kill a process:
channel = await developer.make_channel('com.apple.instruments.server.services.processcontrol')
args = MessageAux().append_obj(80) # This will kill pid 80
await channel.killPid_(args, expects_reply=False) # Killing a process doesn't require an answer.

# In some rare cases, you might want to receive the auxiliary and the selector return value.
# For that cases you can use the recv_plist method.
return_value, auxiliary = await developer.recv_plist()
"""


class BplitAdapter(Adapter):
    def _decode(self, obj, context, path):
        return archiver.unarchive(obj)

    def _encode(self, obj, context, path):
        return archiver.archive(obj)


message_aux_t_struct = Struct(
    "magic" / Default(Int64ul, 0x1F0),
    "aux"
    / Prefixed(
        Int64ul,
        GreedyRange(
            Struct(
                "_empty_dictionary" / Select(Const(0xA, Int32ul), Int32ul),
                "type" / Int32ul,
                "value"
                / Switch(
                    this.type,
                    {2: BplitAdapter(Prefixed(Int32ul, GreedyBytes)), 3: Int32ul, 6: Int64ul},
                    default=GreedyBytes,
                ),
            )
        ),
    ),
)
dtx_message_header_struct = Struct(
    "magic" / Const(0x1F3D5B79, Int32ul),
    "cb" / Int32ul,
    "fragmentId" / Int16ul,
    "fragmentCount" / Int16ul,
    "length" / Int32ul,
    "identifier" / Int32ul,
    "conversationIndex" / Int32ul,
    "channelCode" / Int32sl,
    "expectsReply" / Int32ul,
)
dtx_message_payload_header_struct = Struct(
    "flags" / Int32ul,
    "auxiliaryLength" / Int32ul,
    "totalLength" / Int64ul,
)


class MessageAux:
    def __init__(self):
        self.values = []

    def append_int(self, value: int):
        self.values.append({"type": 3, "value": value})
        return self

    def append_long(self, value: int):
        self.values.append({"type": 6, "value": value})
        return self

    def append_obj(self, value):
        self.values.append({"type": 2, "value": value})
        return self

    def __bytes__(self):
        return message_aux_t_struct.build({"aux": self.values})


class DTTapMessage:
    @staticmethod
    def decode_archive(archive_obj):
        return archive_obj.decode("DTTapMessagePlist")


class NSNull:
    @staticmethod
    def decode_archive(archive_obj):
        return None


class NSError:
    @staticmethod
    def encode_archive(archive_obj):
        return archiver.archive(archive_obj)

    @staticmethod
    def decode_archive(archive_obj):
        user_info = archive_obj.decode("NSUserInfo")
        if user_info.get("NSLocalizedDescription", "").endswith(" - it does not respond to the selector"):
            raise UnrecognizedSelectorError(user_info)
        raise DvtException(archive_obj.decode("NSUserInfo"))


class NSUUID(uuid.UUID):
    @staticmethod
    def uuid4():
        """Generate a random UUID."""
        return NSUUID(bytes=os.urandom(16))

    def encode_archive(self, archive_obj: archiver.ArchivingObject):
        archive_obj.encode("NS.uuidbytes", self.bytes)

    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return NSUUID(bytes=archive_obj.decode("NS.uuidbytes"))


class NSURL:
    def __init__(self, base, relative):
        self.base = base
        self.relative = relative

    def encode_archive(self, archive_obj: archiver.ArchivingObject):
        archive_obj.encode("NS.base", self.base)
        archive_obj.encode("NS.relative", self.relative)

    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return NSURL(archive_obj.decode("NS.base"), archive_obj.decode("NS.relative"))


class NSValue:
    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return archive_obj.decode("NS.rectval")


class NSMutableData:
    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return archive_obj.decode("NS.data")


class NSMutableString:
    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return archive_obj.decode("NS.string")


class XCTCapabilities:
    def __init__(self, capabilities: dict):
        self.capabilities = capabilities

    def encode_archive(self, archive_obj: archiver.ArchivingObject):
        archive_obj.encode("capabilities-dictionary", self.capabilities)

    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return XCTCapabilities(archive_obj.decode("capabilities-dictionary"))

    def __str__(self):
        return f"XCTCapabilities({self.capabilities})"


class XCTestConfiguration:
    _default: ClassVar = {
        # 'testBundleURL': UID(3),
        # 'sessionIdentifier': UID(8), # UUID
        "aggregateStatisticsBeforeCrash": {"XCSuiteRecordsKey": {}},
        "automationFrameworkPath": "/Developer/Library/PrivateFrameworks/XCTAutomationSupport.framework",
        "baselineFileRelativePath": None,
        "baselineFileURL": None,
        "defaultTestExecutionTimeAllowance": None,
        "disablePerformanceMetrics": False,
        "emitOSLogs": False,
        "formatVersion": plistlib.UID(2),  # store in UID
        "gatherLocalizableStringsData": False,
        "initializeForUITesting": True,
        "maximumTestExecutionTimeAllowance": None,
        "productModuleName": "WebDriverAgentRunner",  # set to other value is also OK
        "randomExecutionOrderingSeed": None,
        "reportActivities": True,
        "reportResultsToIDE": True,
        "systemAttachmentLifetime": 2,
        "targetApplicationArguments": [],  # maybe useless
        "targetApplicationBundleID": None,
        "targetApplicationEnvironment": None,
        "targetApplicationPath": "/whatever-it-does-not-matter/but-should-not-be-empty",
        "testApplicationDependencies": {},
        "testApplicationUserOverrides": None,
        "testBundleRelativePath": None,
        "testExecutionOrdering": 0,
        "testTimeoutsEnabled": False,
        "testsDrivenByIDE": False,
        "testsMustRunOnMainThread": True,
        "testsToRun": None,
        "testsToSkip": None,
        "treatMissingBaselinesAsFailures": False,
        "userAttachmentLifetime": 0,
        "preferredScreenCaptureFormat": 2,
        "IDECapabilities": XCTCapabilities({
            "expected failure test capability": True,
            "test case run configurations": True,
            "test timeout capability": True,
            "test iterations": True,
            "request diagnostics for specific devices": True,
            "delayed attachment transfer": True,
            "skipped test capability": True,
            "daemon container sandbox extension": True,
            "ubiquitous test identifiers": True,
            "XCTIssue capability": True,
        }),
    }

    def __init__(self, kv: dict):
        assert "testBundleURL" in kv
        assert "sessionIdentifier" in kv
        self._config = copy.deepcopy(self._default)
        self._config.update(kv)

    def encode_archive(self, archive_obj: archiver.ArchivingObject):
        for k, v in self._config.items():
            archive_obj.encode(k, v)

    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return archive_obj.object


archiver.update_class_map({
    "DTSysmonTapMessage": DTTapMessage,
    "DTTapHeartbeatMessage": DTTapMessage,
    "DTTapStatusMessage": DTTapMessage,
    "DTKTraceTapMessage": DTTapMessage,
    "DTActivityTraceTapMessage": DTTapMessage,
    "DTTapMessage": DTTapMessage,
    "NSNull": NSNull,
    "NSError": NSError,
    "NSUUID": NSUUID,
    "NSURL": NSURL,
    "NSValue": NSValue,
    "NSMutableData": NSMutableData,
    "NSMutableString": NSMutableString,
    "XCTestConfiguration": XCTestConfiguration,
    "XCTCapabilities": XCTCapabilities,
})

archiver.Archive.inline_types = list({*archiver.Archive.inline_types, bytes})


class Channel(int):
    @classmethod
    def create(cls, value: int, service: "RemoteServer"):
        """Attach a `RemoteServer` instance to an integer channel code."""
        channel = cls(value)
        channel._service = service
        return channel

    async def receive_key_value(self):
        """Receive a `(selector, aux)` tuple from this channel."""
        return await self._service.recv_plist(self)

    async def receive_plist(self):
        """Receive and return only the decoded selector/payload object."""
        return (await self._service.recv_plist(self))[0]

    async def receive_message(self):
        """Receive raw decoded message payload bytes and aux values."""
        return (await self._service.recv_message(self))[0]

    async def send_message(self, selector: str, args: MessageAux = None, expects_reply: bool = True):
        """Send a selector call on this channel."""
        await self._service.send_message(self, selector, args, expects_reply=expects_reply)

    @staticmethod
    def _sanitize_name(name: str):
        """
        Sanitize python name to ObjectiveC name.
        """
        name = "_" + name[1:].replace("_", ":") if name.startswith("_") else name.replace("_", ":")
        return name

    def __getitem__(self, item) -> Callable[[MessageAux], Awaitable[Any]]:
        """Return a callable proxy for sending `item` as selector."""
        return partial(self._service.send_message, self, item)

    def __getattr__(self, item) -> Callable[[MessageAux], Awaitable[Any]]:
        """Resolve unknown attributes to Objective-C selector proxies."""
        return self[self._sanitize_name(item)]


class RemoteServer(LockdownService):
    """
    Wrapper to Apple's RemoteServer.
    This server exports several ObjC objects allowing calling their respective selectors.
    The `/Developer/Library/PrivateFrameworks/DVTInstrumentsFoundation.framework/DTServiceHub` service reads the
    configuration stored from `[[NSUserDefaults standardUserDefaults] boolForKey:@"DTXConnectionTracer"]`
    If the value is true, then `/tmp/DTServiceHub[PID].DTXConnection.RANDOM.log` is created and can be used to debug the
    transport protocol.

    For example:

    ```
    root@iPhone (/var/root)# tail -f /tmp/DTServiceHub[369].DTXConnection.qNjM2U.log
    170.887982 x4 resuming [c0]: <DTXConnection 0x100d20670 : x4>
    170.889120 x4   sent   [c0]: < DTXMessage 0x100d52b10 : i2.0 c0 dispatch:[_notifyOfPublishedCapabilities:<NSDictionary 0x100d0e1b0 | 92 key/value pairs>] >
    170.889547 x4 received [c0]: < DTXMessage 0x100d0a550 : i1.0 c0 dispatch:[_notifyOfPublishedCapabilities:<NSDictionary 0x100d16a40 | 2 key/value pairs>] >
    170.892101 x4 received [c0]: < DTXMessage 0x100d0a550 : i3.0e c0 dispatch:[_requestChannelWithCode:[1]identifier :"com.apple.instruments.server.services.deviceinfo"] >
    170.892238 x4   sent   [c0]: < DTXMessage 0x100d61830 : i3.1 c0 >
    170.892973 x4 received [c1f]: < DTXMessage 0x100d0a550 : i4.0e c1 dispatch:[runningProcesses] >
    171.204957 x4   sent   [c1f]: < DTXMessage 0x100c557a0 : i4.1 c1 object:(__NSArrayM*)<NSArray 0x100c199d0 | 245 objects> { <NSDictionary 0x100c167c0 | 5 key/value pairs>, <NSDictionary 0x100d17970 | 5 key/value pairs>, <NSDictionary 0x100d17f40 | 5 key/value pairs>, <NSDictionary 0x100d61750 | 5 key/value pairs>, <NSDictionary 0x100c16760 | 5 key/value pairs>, ...  } >
    171.213326 x4 received [c0]: < DTXMessage : kDTXInterruptionMessage >
    171.213424 x4  handler [c0]: < DTXMessage : i1 kDTXInterruptionMessage >
    171.213477 x4 received [c1f]: < DTXMessage : kDTXInterruptionMessage >
    ```

    For editing the configuration we can simply add the respected key into:
    `/var/mobile/Library/Preferences/.GlobalPreferences.plist` and kill `cfprefsd`

    The valid selectors for triggering can be found using the following Frida script the same way Troy Bowman used for
    iterating all classes which implement the protocol `DTXAllowedRPC`:

    ```shell
    frida -U DTServiceHub
    ```

    ```javascript
    for (var name in ObjC.protocols) {
        var protocol = ObjC.protocols[name]
        if ('DTXAllowedRPC' in protocol.protocols) {
            console.log('@protocol', name)
            console.log('  ' + Object.keys(protocol.methods).join('\n  '))
        }
    }
    ```

    Lifecycle overview:
    - `connect()` establishes the transport, optionally negotiates SSL, and starts `_reader_loop()`.
    - `_reader_loop()` continuously reads frames, reassembles fragments, and enqueues complete messages by channel.
    - API calls (`send_message`, `recv_message`, `recv_plist`) operate on those per-channel queues.
    - `make_channel()` creates and caches channels with single-flight dedup for concurrent callers.
    - `close()` sends `_channelCanceled:`, stops the reader, and closes the underlying lockdown service.
    """

    BROADCAST_CHANNEL = 0
    INSTRUMENTS_MESSAGE_TYPE = 2
    CHANNEL_CLOSED = (None, b"")

    def __init__(
        self,
        lockdown: LockdownServiceProvider,
        service_name,
        remove_ssl_context: bool = True,
        is_developer_service: bool = True,
    ):
        """Initialize connection state, channel registries, and synchronization primitives."""
        super().__init__(lockdown, service_name, is_developer_service=is_developer_service)
        self._remove_ssl_context = remove_ssl_context

        self.supported_identifiers = {}

        # channel_cache, _pending_channel_requests, last_channel_code, channel_messages
        # work together to ensure only one active creator for each channel identifier.
        # They are to be used only within the protection of _channel_lock.
        # In any given moment, for a given channel identifier:
        # - channel_cache contains its channel OR
        # - _pending_channel_requests contains a Future for its creation OR
        # - neither if it has not been requested yet.
        # channel_messages is populated on demand ( on first read or write ),
        # a value of (None, b"") is enqueued when the channel is closed.
        self._channel_lock = asyncio.Lock()
        self.channel_cache: dict[str, Channel] = {}
        self._pending_channel_requests: dict[str, asyncio.Future] = {}
        self.last_channel_code = 0
        self.channel_messages = {self.BROADCAST_CHANNEL: asyncio.Queue()}

        self._fragment_buffers: dict[tuple[int, int, int], bytes] = {}

        # cur_message holds the identifier of the sent messages.
        # It is independent of the identifier of the received messages.
        # Send operations ( self.service.sendall ) and message counting (`self.cur_message++`)
        # must be protected by _send_lock to avoid concurrent writing on the socket.
        self._send_lock = asyncio.Lock()
        self.cur_message = 0

        # reader_task holds the background task responsible for reading from the socket,
        # defragmenting messages, and dispatching them to channel queues.
        # Ensure its existence using _ensure_reader() before any operation that expects incoming messages
        # and check _reader_closed to detect a closed connection.
        self._reader_lock = asyncio.Lock()
        self._reader_task: Optional[asyncio.Task] = None
        self._reader_closed = False

        self.broadcast = Channel.create(0, self)

    async def _get_channel_queue(self, channel: int) -> asyncio.Queue:
        """Return (or lazily create) the inbound queue for a given channel code."""
        async with self._channel_lock:
            if channel not in self.channel_messages:
                self.channel_messages[channel] = asyncio.Queue()
            return self.channel_messages[channel]

    async def _ensure_reader(self) -> None:
        """Start the background reader task if it is not currently running."""
        async with self._reader_lock:
            if self._reader_closed:
                raise ConnectionTerminatedError()
            if self._reader_task is None:
                self.logger.debug("starting reader task")
                self._reader_task = asyncio.create_task(self._reader_loop(), name="remote-server-reader")
                return
            if not self._reader_task.done():
                return  # reader running
            try:
                e = self._reader_task.exception()
            except asyncio.CancelledError as e1:
                e = e1
            self.logger.exception("reader task exited unexpectedly", exc_info=e)
            raise ConnectionTerminatedError() from e

    async def _stop_reader(self) -> None:
        """Cancel and await the background reader task if present."""
        async with self._reader_lock:
            if self._reader_closed:
                return
            self._reader_closed = True
            t = self._reader_task
        if t is not None:
            self.logger.debug("stopping reader task")
            t.cancel()
            with suppress(asyncio.CancelledError):
                await t

    async def _reader_loop(self) -> None:
        """
        Continuously read raw DTX frames, defragment by message key, and fan out complete
        payloads into per-channel queues.

        Notes:
        - For fragmented messages, the first frame (fragmentId=0) carries only the DTX
          message header. Subsequent fragments carry payload bytes.
        - Fragment assembly is keyed by ``(identifier, conversationIndex, abs(channelCode))``
          to keep request/reply and per-channel streams isolated.
        - Once the final fragment arrives, a normalized single-fragment header and full
          payload are pushed to the channel queue.
        - On exit (error or cancellation), a ``(None, b"")`` sentinel is sent to every
          channel queue so awaiters can terminate promptly.
        """
        try:
            # FIXME: pthread_cond_wait like on the _reader_lock + _reader_closed c
            while True:
                data = await self.service.recvall(dtx_message_header_struct.sizeof())
                mheader = dtx_message_header_struct.parse(data)

                if mheader.fragmentCount > 1 and mheader.fragmentId == 0:
                    # when reading multiple message fragments, the first fragment contains only a message header
                    continue

                chunk = await self.service.recvall(mheader.length)

                # treat both as the negative and positive representation of the channel code in the response
                # the same when performing fragmentation
                received_channel_code = abs(mheader.channelCode)
                fragment_key = (mheader.identifier, mheader.conversationIndex, received_channel_code)
                payload = self._fragment_buffers.get(fragment_key, b"") + chunk
                if mheader.fragmentId == mheader.fragmentCount - 1:
                    self._fragment_buffers.pop(fragment_key, None)
                    assembled_header = mheader.copy()
                    assembled_header.fragmentId = 0
                    assembled_header.fragmentCount = 1
                    assembled_header.channelCode = received_channel_code
                    assembled_header.length = len(payload)
                    self._log_dtx_message("received", assembled_header, payload)
                    (await self._get_channel_queue(received_channel_code)).put_nowait((assembled_header, payload))
                else:
                    self._fragment_buffers[fragment_key] = payload
        except ConnectionResetError:
            self.logger.info("connection reset by peer, stopping reader loop")
            async with self._reader_lock:
                self._reader_closed = True
        finally:
            self.logger.debug("reader loop exiting, closing all channels")
            async with self._channel_lock:
                for q in self.channel_messages.values():
                    q.put_nowait(self.CHANNEL_CLOSED)

    async def connect(self) -> None:
        """Connect transport, complete protocol prerequisites, and ensure reader loop is active."""
        if self._service is None and self._remove_ssl_context:
            attr = await self.lockdown.get_service_connection_attributes(
                self.service_name, include_escrow_bag=self._include_escrow_bag
            )
            self._service = await self.lockdown.create_service_connection(attr["Port"])
            if attr.get("EnableServiceSSL", False) and hasattr(self.lockdown, "ssl_file"):
                # Mirror the legacy sync flow: negotiate SSL once, then strip SSL context for raw DTX traffic.
                # This is currently needed for RemoteServer services (DTX protocol) which only perform the
                # handshake in TLS. This happens for:
                # - AccessibilityAudit service
                # - iOS < 14.0
                with self.lockdown.ssl_file() as f:
                    self._service.setblocking(True)
                    self._service.ssl_start_sync(f)
        else:
            await super().connect()
        if self._remove_ssl_context:
            await self._recreate_connection_without_ssl_context()
        await self._ensure_reader()

    async def _recreate_connection_without_ssl_context(self) -> None:
        """Rebind to the underlying raw socket after TLS bootstrap to keep async stream mode."""
        if self._service is None or self._service.socket is None:
            return

        ssl_socket = self._service.socket
        if not hasattr(ssl_socket, "_sslobj"):
            return

        raw_socket = getattr(ssl_socket, "_sock", None)
        if raw_socket is None:
            # Python 3.14 may not expose ``_sock`` on SSLSocket. Detach the FD
            # and recreate a plain socket so asyncio can attach stream transports.
            fd = ssl_socket.detach()
            raw_socket = socket.socket(fileno=fd)
        else:
            ssl_socket._sslobj = None

        self._service.socket = raw_socket
        self._service.reader = None
        self._service.writer = None
        self._service.socket.setblocking(False)
        await self._service.start()

    def _log_dtx_message(self, direction: str, mheader: Union[Container, bytes], payload: bytes) -> None:
        """Best-effort DEBUG log formatter for DTX packets."""
        if not self.logger.isEnabledFor(logging.DEBUG):
            return

        if isinstance(mheader, bytes):
            mheader = dtx_message_header_struct.parse(mheader)

        s = io.BytesIO(payload)
        pheader = dtx_message_payload_header_struct.parse_stream(s)
        compression = pheader.flags == 0x0707
        if compression:
            raise NotImplementedError("Compressed")

        aux = message_aux_t_struct.parse_stream(s).aux if pheader.auxiliaryLength else None
        obj_size = pheader.totalLength - pheader.auxiliaryLength
        data = s.read(obj_size) if obj_size else None

        with suppress(Exception):
            if pheader.auxiliaryLength:
                aux = [c.value for c in aux]
        with suppress(Exception):
            if obj_size:
                data = archiver.unarchive(data)

        e = "e" if mheader.expectsReply else ""
        self.logger.debug(
            "x%d %-8s [c?]: <DTXMessage: i%d.%d%s c%d t:%d payload:%s aux:%s>",
            self.service.socket.fileno(),
            direction,
            mheader.identifier,
            mheader.conversationIndex,
            e,
            mheader.channelCode,
            pheader.flags,
            data,
            aux,
        )

    def shell(self):
        """Launch an IPython shell preloaded with this server and its broadcast channel."""
        start_ipython_shell(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.Terminal256Formatter(style="native")),
            user_ns={
                "developer": self,
                "broadcast": self.broadcast,
                "MessageAux": MessageAux,
            },
        )

    async def perform_handshake(self):
        """Exchange published-capabilities messages and cache remote supported identifiers."""
        args = MessageAux()
        args.append_obj({"com.apple.private.DTXBlockCompression": 0, "com.apple.private.DTXConnection": 1})

        await self.send_message(0, "_notifyOfPublishedCapabilities:", args, expects_reply=False)
        ret, aux = await self.recv_plist()
        if ret != "_notifyOfPublishedCapabilities:":
            raise ValueError("Invalid answer")
        if not len(aux[0]):
            raise ValueError("Invalid answer")
        self.supported_identifiers = aux[0].value

    async def make_channel(self, identifier: str) -> Channel:
        """
        Get or create a DTX channel for a service identifier.

        Concurrency model:
        - Fast path: return cached channel if already created.
        - Dedup path: if another coroutine is already creating the same identifier,
          await its shared Future instead of issuing another request.
        - Create path: only one coroutine sends ``_requestChannelWithCode:identifier:``
          and later publishes the created channel into ``channel_cache`` while resolving
          the shared Future for waiters.

        The ``_state_lock`` protects all shared mutable state involved in the above:
        ``channel_cache``, ``_pending_channel_requests``, and ``last_channel_code``.
        """
        # NOTE: There is also identifier not in self.supported_identifiers
        # assert identifier in self.supported_identifiers
        async with self._channel_lock:
            existing = self.channel_cache.get(identifier)
            if existing is not None:
                return existing

            pending_request = self._pending_channel_requests.get(identifier)
            if pending_request is None:
                pending_request = asyncio.get_running_loop().create_future()
                self._pending_channel_requests[identifier] = pending_request
                self.last_channel_code += 1
                code = self.last_channel_code
                create_channel = True
            else:
                create_channel = False

        if not create_channel:
            # Another coroutine already owns channel creation for this identifier.
            return await pending_request

        try:
            args = MessageAux().append_int(code).append_obj(identifier)
            await self.send_message(0, "_requestChannelWithCode:identifier:", args)
            ret, _aux = await self.recv_plist()
            assert ret is None
            created_channel = Channel.create(code, self)
            async with self._channel_lock:
                self.channel_cache[identifier] = created_channel
                self._pending_channel_requests.pop(identifier, None)
                pending_request.set_result(created_channel)
        except BaseException as e:
            async with self._channel_lock:
                self.channel_cache.pop(identifier, None)
                self._pending_channel_requests.pop(identifier, None)
                pending_request.set_exception(e)
            raise
        return created_channel

    async def serve_channel(self, identifier: str, channel: Optional[Channel] = None) -> Channel:
        """Accept a reverse-channel request from peer and acknowledge it with an ACK reply."""
        msg, mheader = await self.recv_plist(self.BROADCAST_CHANNEL, return_header=True)
        assert msg[0] == "_requestChannelWithCode:identifier:", f"expected a request for a reverse channel, got: {msg}"
        code = msg[1][0].value
        assert channel is None or channel == code, (
            f"expected a request for a reverse channel with channelCode:{channel}, got:{msg[1][0]}"
        )
        received_identifier = msg[1][1].value
        assert identifier == received_identifier, (
            f"expected a request for a reverse channel with identifier:{identifier}, got:{received_identifier}"
        )
        if channel is None:
            channel = Channel.create(code, self)
        await self.send_reply_ack(self.BROADCAST_CHANNEL, mheader.identifier, None, None)
        return channel

    async def send_message(
        self, channel: int, selector: Optional[str] = None, args: MessageAux = None, expects_reply: bool = True
    ):
        """Serialize and send a DTX method invocation to a specific channel."""
        aux = bytes(args) if args is not None else b""
        sel = archiver.archive(selector) if selector is not None else b""
        flags = self.INSTRUMENTS_MESSAGE_TYPE
        pheader = dtx_message_payload_header_struct.build({
            "flags": flags,
            "auxiliaryLength": len(aux),
            "totalLength": len(aux) + len(sel),
        })

        async with self._send_lock:
            self.cur_message += 1
            mheader = dtx_message_header_struct.build({
                "cb": dtx_message_header_struct.sizeof(),
                "fragmentId": 0,
                "fragmentCount": 1,
                "length": dtx_message_payload_header_struct.sizeof() + len(aux) + len(sel),
                "identifier": self.cur_message,
                "conversationIndex": 0,
                "channelCode": channel,
                "expectsReply": int(expects_reply),
            })
            msg = mheader + pheader + aux + sel
            await self.service.sendall(msg)
            self._log_dtx_message("sent", mheader, pheader + aux + sel)

    async def recv_plist(self, channel: int = BROADCAST_CHANNEL, return_header: bool = False):
        """Receive a channel message and decode plist/archived object payloads."""
        recv_result = await self.recv_message(channel, return_header=return_header)
        if return_header:
            data, aux, mheader = recv_result
        else:
            data, aux = recv_result
        if data is not None:
            try:
                data = archiver.unarchive(data)
            except archiver.MissingClassMapping:
                pprint(plistlib.loads(data))
                raise
            except plistlib.InvalidFileException:
                self.logger.warning(f"got an invalid plist: {data[:40]}")
        if return_header:
            return (data, aux), mheader
        return data, aux

    async def recv_message(self, channel: int = BROADCAST_CHANNEL, return_header: bool = False):
        """Receive one complete DTX payload from a channel queue and parse its framing."""
        mheader, packet_stream = await self._recv_packet_fragments(channel)
        pheader = dtx_message_payload_header_struct.parse_stream(packet_stream)

        compression = (pheader.flags & 0xFF000) >> 12
        if compression:
            raise NotImplementedError("Compressed")

        aux = message_aux_t_struct.parse_stream(packet_stream).aux if pheader.auxiliaryLength else None
        obj_size = pheader.totalLength - pheader.auxiliaryLength
        data = packet_stream.read(obj_size) if obj_size else None
        if return_header:
            return data, aux, mheader
        return data, aux

    # TODO: rewrite the RemoteServer class (possibly even ServiceConnection) to continue consume messages
    # TODO: improve reply correlation by tracking expected replies with message IDs, rather than relying on channel-ordered consumers.

    async def _send_reply(
        self,
        channel: int,
        msg_id: int,
        msg_type: int,
        payload: Optional[object] = None,
        aux: Optional[MessageAux] = None,
    ):
        """Low-level DTX reply sender used by typed reply helpers."""
        payload_bytes = archiver.archive(payload) if payload is not None else b""
        aux_bytes = archiver.archive(aux) if aux is not None else b""
        pheader = dtx_message_payload_header_struct.build({
            "flags": msg_type,
            "auxiliaryLength": len(aux_bytes),
            "totalLength": len(payload_bytes) + len(aux_bytes),
        })
        mheader = dtx_message_header_struct.build({
            "cb": dtx_message_header_struct.sizeof(),
            "fragmentId": 0,
            "fragmentCount": 1,
            "length": dtx_message_payload_header_struct.sizeof() + len(payload_bytes) + len(aux_bytes),
            "identifier": msg_id,
            "conversationIndex": 1,
            "channelCode": channel,
            "expectsReply": (0),
        })
        msg = mheader + pheader + aux_bytes + payload_bytes
        async with self._send_lock:
            await self.service.sendall(msg)
            self._log_dtx_message("sent", mheader, pheader + aux_bytes + payload_bytes)

    async def send_reply(
        self, channel: int, msg_id: int, payload: Optional[object] = None, aux: Optional[MessageAux] = None
    ):
        """Send a success reply (`ResponseWithReturnValueInPayload`) for a request."""
        msg_type = 0x3  # ResponseWithReturnValueInPayload
        await self._send_reply(channel, msg_id, msg_type, payload, aux)

    async def send_reply_error(
        self, channel: int, msg_id: int, payload: Optional[object] = None, aux: Optional[MessageAux] = None
    ):
        """Send an error reply (`DtxTypeError`) for a request."""
        msg_type = 0x4  # DtxTypeError
        await self._send_reply(channel, msg_id, msg_type, payload, aux)

    async def send_reply_ack(
        self, channel: int, msg_id: int, payload: Optional[object] = None, aux: Optional[MessageAux] = None
    ):
        """Send an ACK-only reply for a request."""
        msg_type = 0x0  # Ack
        await self._send_reply(channel, msg_id, msg_type, payload, aux)

    async def _recv_packet_fragments(self, channel: int = BROADCAST_CHANNEL):
        """Await next fully-assembled packet for `channel` from the reader loop."""
        # here the order matters: get the channel and then check for the reader
        queue = await self._get_channel_queue(channel)
        await self._ensure_reader()

        mheader, message = await queue.get()

        if (mheader, message) == self.CHANNEL_CLOSED:
            await self._ensure_reader()  # this will raise ConnectionTerminatedError if the reader is closed
            raise ChannelClosedError()  # otherwise the other end genuinely requested a channel close
        return mheader, io.BytesIO(message)

    def __enter__(self):
        """Synchronous context manager is intentionally unsupported."""
        raise RuntimeError("Use async context manager: `async with ...`")

    async def close(self):
        """Gracefully shut down channels, stop reader loop, and close underlying service."""
        async with self._channel_lock:
            codes = [code for code in self.channel_messages if code > 0]
        if codes:
            aux = MessageAux()
            for code in codes:
                aux.append_int(code)

            try:
                await self.send_message(self.BROADCAST_CHANNEL, "_channelCanceled:", aux, expects_reply=False)
            except OSError:
                # ignore: OSError: [Errno 9] Bad file descriptor
                pass
            except RuntimeError as e:
                # Async generator teardown / interrupted CLI flows may close without a running loop.
                if "no running event loop" not in str(e):
                    raise

        await self._stop_reader()
        await super().close()

    async def __aenter__(self):
        """Connect, handshake, and return this instance for async context manager use."""
        await self.connect()
        await self.perform_handshake()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Ensure graceful shutdown on async context manager exit."""
        await self.close()


class ChannelService:
    """Lazy channel helper for DVT/RemoteServer-backed services."""

    IDENTIFIER: str

    def __init__(self, dvt: "RemoteServer", channel_name: Optional[str] = None) -> None:
        """Store server reference and optional channel identifier override."""
        self._dvt = dvt
        self._channel_name = channel_name if channel_name is not None else self.IDENTIFIER
        self._channel = None

    async def _channel_ref(self) -> Channel:
        """Lazily create and cache the backing channel."""
        if self._channel is None:
            self._channel = await self._dvt.make_channel(self._channel_name)
        return self._channel


class Tap(ChannelService):
    def __init__(self, dvt, channel_name: str, config: dict) -> None:
        """Initialize a tap helper with channel name and tap configuration."""
        super().__init__(dvt, channel_name=channel_name)
        self._config = config
        self.channel = None

    def __enter__(self):
        """Synchronous context manager is intentionally unsupported."""
        raise RuntimeError("Use async context manager: `async with ...`")

    async def __aenter__(self):
        """Open channel, configure tap, and start streaming."""
        self.channel = await self._channel_ref()
        await self.channel.setConfig_(MessageAux().append_obj(self._config), expects_reply=False)
        await self.channel.start(expects_reply=False)
        # first message is just kind of an ack
        await self.channel.receive_plist()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Stop tap streaming when leaving async context manager."""
        if self.channel is not None:
            await self.channel.stop(expects_reply=False)

    async def __aiter__(self):
        """Yield messages continuously from the active tap stream."""
        while True:
            for message in await self.channel.receive_plist():
                yield message



import asyncio
import contextlib
import logging
import plistlib
import socket
import ssl
import struct
import time
import xml
from enum import Enum
from typing import Any, Optional, Union

from pygments import formatters, highlight, lexers

from pymobiledevice3.exceptions import (
    ConnectionTerminatedError,
    DeviceNotFoundError,
    NoDeviceConnectedError,
    PyMobileDevice3Exception,
)
from pymobiledevice3.osu.os_utils import get_os_utils
from pymobiledevice3.usbmux import MuxDevice, select_device
from pymobiledevice3.utils import start_ipython_shell

DEFAULT_AFTER_IDLE_SEC = 3
DEFAULT_INTERVAL_SEC = 3
DEFAULT_MAX_FAILS = 3
DEFAULT_TIMEOUT = 1
DEFAULT_SSL_HANDSHAKE_TIMEOUT = 10
OSUTIL = get_os_utils()
SHELL_USAGE = """
# This shell allows you to communicate directly with every service layer behind the lockdownd daemon.

# For example, you can do the following:
await client.send_plist({"Command": "DoSomething"})

# and view the reply
print(await client.recv_plist())

# or just send raw message
client.send(b"hello")

# and view the result
print(await client.recvall(20))
"""


def build_plist(d: dict, endianity: str = ">", fmt: Enum = plistlib.FMT_XML) -> bytes:
    """
    Convert a dictionary to a plist-formatted byte string prefixed with a length field.

    :param d: The dictionary to convert.
    :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
    :param fmt: The plist format (e.g., plistlib.FMT_XML).
    :return: The plist-formatted byte string.
    """
    payload = plistlib.dumps(d, fmt=fmt)
    message = struct.pack(endianity + "L", len(payload))
    return message + payload


def parse_plist(payload: bytes) -> dict:
    """
    Parse a plist-formatted byte string into a dictionary.

    :param payload: The plist-formatted byte string to parse.
    :return: The parsed dictionary.
    :raises PyMobileDevice3Exception: If the payload is invalid.
    :retries with a filtered payload of only valid XML characters if plistlib compains about "not well-formed (invalid token)"
    """
    try:
        return plistlib.loads(payload)
    except xml.parsers.expat.ExpatError:
        payload = bytes([b for b in payload if b >= 0x20 or b in (0x09, 0x0A, 0x0D)])
        try:
            return plistlib.loads(payload)
        except plistlib.InvalidFileException as e:
            raise PyMobileDevice3Exception(f"parse_plist invalid data: {payload[:100].hex()}") from e
    except plistlib.InvalidFileException as e:
        raise PyMobileDevice3Exception(f"parse_plist invalid data: {payload[:100].hex()}") from e


class ServiceConnection:
    """wrapper for tcp-relay connections"""

    def __init__(self, sock: socket.socket, mux_device: Optional[MuxDevice] = None) -> None:
        """
        Initialize a ServiceConnection object.

        :param sock: The socket to use for the connection.
        :param mux_device: The MuxDevice associated with the connection (optional).
        """
        self.logger = logging.getLogger(__name__)
        self.socket = sock
        self._offset = 0

        # usbmux connections contain additional information associated with the current connection
        self.mux_device = mux_device

        # Async stream reader and writer used for stream mode.
        self.reader = None  # type: Optional[asyncio.StreamReader]
        self.writer = None  # type: Optional[asyncio.StreamWriter]

        # SSL/TLS version to be used for connecting to device
        # TLS v1.2 is supported since iOS 5
        self.min_ssl_proto = ssl.TLSVersion.TLSv1_2
        self.max_ssl_proto = ssl.TLSVersion.TLSv1_3
        self.socket.setblocking(False)

    @staticmethod
    async def create_using_tcp(
        hostname: str, port: int, keep_alive: bool = True, create_connection_timeout: int = DEFAULT_TIMEOUT
    ) -> "ServiceConnection":
        """
        Create a ServiceConnection using a TCP connection.

        :param hostname: The hostname of the server to connect to.
        :param port: The port to connect to.
        :param keep_alive: Whether to enable TCP keep-alive.
        :param create_connection_timeout: The timeout for creating the connection.
        :return: A ServiceConnection object.
        """
        reader, writer = await asyncio.wait_for(
            asyncio.open_connection(hostname, port), timeout=create_connection_timeout
        )
        sock = writer.get_extra_info("socket")
        if sock is None:
            writer.close()
            with contextlib.suppress(Exception):
                await writer.wait_closed()
            raise ConnectionError(f"failed to get socket from connection to {hostname}:{port}")
        if keep_alive:
            OSUTIL.set_keepalive(sock)
        conn = ServiceConnection(sock)
        conn.reader = reader
        conn.writer = writer
        return conn

    @staticmethod
    async def create_using_usbmux(
        udid: Optional[str], port: int, connection_type: Optional[str] = None, usbmux_address: Optional[str] = None
    ) -> "ServiceConnection":
        """
        Create a ServiceConnection through usbmuxd.

        :param udid: Target device UDID (or None for first available matching device).
        :param port: Device-side service port to connect to.
        :param connection_type: Optional transport filter (e.g. ``USB``/``Network``).
        :param usbmux_address: Optional override for the usbmuxd socket address.
        :return: A ServiceConnection bound to the selected device/port.
        :raises DeviceNotFoundError: If the requested UDID is not found.
        :raises NoDeviceConnectedError: If no matching devices are connected.
        """
        target_device = await select_device(udid, connection_type=connection_type, usbmux_address=usbmux_address)
        if target_device is None:
            if udid:
                raise DeviceNotFoundError(udid)
            raise NoDeviceConnectedError()
        sock = await target_device.connect(port, usbmux_address=usbmux_address)
        return ServiceConnection(sock, mux_device=target_device)

    def setblocking(self, blocking: bool) -> None:
        """
        Set the blocking mode of the socket.

        :param blocking: If True, set the socket to blocking mode; otherwise, set it to non-blocking mode.
        """
        self.socket.setblocking(blocking)

    async def _ensure_started(self) -> None:
        if self.reader is not None and self.writer is not None:
            return
        await self.start()

    async def aclose(self) -> None:
        """Asynchronously close the connection."""
        await self.close()

    async def close(self) -> None:
        """Asynchronously close the connection."""
        if self.writer is not None:
            with contextlib.suppress(Exception):
                self.writer.close()
            with contextlib.suppress(Exception):
                await self.writer.wait_closed()
            if self.socket is not None and self.socket.fileno() != -1:
                with contextlib.suppress(Exception):
                    self.socket.close()
        elif self.socket is not None:
            with contextlib.suppress(Exception):
                self.socket.close()
        self.socket = None
        self.writer = None
        self.reader = None

    def recv_sync(self, length: int = 4096) -> bytes:
        """
        Receive data from the socket.

        :param length: The maximum amount of data to receive.
        :return: The received data.
        """
        try:
            return self.socket.recv(length)
        except (ssl.SSLError, BrokenPipeError) as e:
            raise ConnectionTerminatedError() from e

    async def recv_any(self, length: int = 4096) -> bytes:
        """
        Asynchronously receive up to ``length`` bytes from the socket/stream.
        """
        await self._ensure_started()
        return await self.reader.read(length)

    def sendall_sync(self, data: bytes) -> None:
        """
        Send data to the socket.

        :param data: The data to send.
        :raises ConnectionTerminatedError: If the connection is terminated abruptly.
        """
        try:
            self.socket.sendall(data)
        except ssl.SSLEOFError as e:
            raise ConnectionTerminatedError from e

    async def send_recv_plist(self, data: dict, endianity: str = ">", fmt: Enum = plistlib.FMT_XML) -> Any:
        """
        Asynchronously send a plist to the socket and receive a plist response.

        :param data: The dictionary to send as a plist.
        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :param fmt: The plist format (e.g., plistlib.FMT_XML).
        :return: The received plist as a dictionary.
        """
        await self.send_plist(data, endianity=endianity, fmt=fmt)
        return await self.recv_plist(endianity=endianity)

    def recvall_sync(self, size: int) -> bytes:
        """
        Receive all data of a specified size from the socket.

        :param size: The amount of data to receive.
        :return: The received data.
        :raises ConnectionTerminatedError: If the connection is aborted.
        """
        data = b""
        while len(data) < size:
            chunk = self.recv_sync(size - len(data))
            if chunk is None or len(chunk) == 0:
                raise ConnectionTerminatedError()
            data += chunk
        return data

    def recv_prefixed_sync(self, endianity: str = ">") -> bytes:
        """
        Receive a data block prefixed with a length field.

        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :return: The received data block.
        """
        size = self.recvall_sync(4)
        if not size or len(size) != 4:
            return b""
        size = struct.unpack(endianity + "L", size)[0]
        while True:
            try:
                return self.recvall_sync(size)
            except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
                # Allow ssl to do stuff
                time.sleep(0)

    async def recvall(self, size: int) -> bytes:
        """
        Asynchronously receive data of a specified size from the socket.

        :param size: The amount of data to receive.
        :return: The received data.
        """
        await self._ensure_started()
        return await self.reader.readexactly(size)

    async def recv_prefixed(self, endianity: str = ">") -> bytes:
        """
        Asynchronously receive a data block prefixed with a length field.

        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :return: The received data block.
        """
        size = await self.recvall(4)
        size = struct.unpack(endianity + "L", size)[0]
        return await self.recvall(size)

    async def send_prefixed(self, data: bytes) -> None:
        """
        Send a data block prefixed with a length field.

        :param data: The data to send.
        """
        if isinstance(data, str):
            data = data.encode()
        hdr = struct.pack(">L", len(data))
        msg = b"".join([hdr, data])
        await self.sendall(msg)

    def recv_plist_sync(self, endianity: str = ">") -> Union[dict, list]:
        """
        Receive a plist from the socket and parse it into a native type.

        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :return: The received plist as a native type.
        """
        return parse_plist(self.recv_prefixed_sync(endianity=endianity))

    async def recv_plist(self, endianity: str = ">") -> dict:
        """
        Asynchronously receive a plist from the socket and parse it into a native type.

        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :return: The received plist as a native type.
        """
        return parse_plist(await self.recv_prefixed(endianity))

    async def sendall(self, payload: bytes) -> None:
        """
        Asynchronously send data to the socket.

        :param payload: The data to send.
        """
        await self._ensure_started()
        try:
            self.writer.write(payload)
            await self.writer.drain()
        except ssl.SSLEOFError as e:
            raise ConnectionTerminatedError from e

    async def send_plist(self, d: Union[dict, list], endianity: str = ">", fmt: Enum = plistlib.FMT_XML) -> None:
        """
        Asynchronously send a dictionary as a plist to the socket.

        :param d: The dictionary to send.
        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :param fmt: The plist format (e.g., plistlib.FMT_XML).
        """
        await self.sendall(build_plist(d, endianity, fmt))

    def create_ssl_context(self, certfile: str, keyfile: Optional[str] = None) -> ssl.SSLContext:
        """
        Create an SSL context for a secure connection.

        :param certfile: The path to the certificate file.
        :param keyfile: The path to the key file (optional).
        :return: An SSL context object.
        """
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.minimum_version = self.min_ssl_proto
        context.maximum_version = self.max_ssl_proto
        if ssl.OPENSSL_VERSION.lower().startswith("openssl"):
            context.set_ciphers("ALL:!aNULL:!eNULL:@SECLEVEL=0")
        else:
            context.set_ciphers("ALL:!aNULL:!eNULL")
        context.options |= 0x4  # OPENSSL OP_LEGACY_SERVER_CONNECT (required for legacy iOS devices)
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        context.load_cert_chain(certfile, keyfile)
        return context

    def ssl_start_sync(self, certfile: str, keyfile: Optional[str] = None) -> None:
        """
        Start an SSL connection.

        :param certfile: The path to the certificate file.
        :param keyfile: The path to the key file (optional).
        """
        try:
            self.socket.settimeout(DEFAULT_SSL_HANDSHAKE_TIMEOUT)
            self.socket = self.create_ssl_context(certfile, keyfile=keyfile).wrap_socket(self.socket)
        except OSError as e:
            raise ConnectionTerminatedError() from e
        finally:
            if self.socket is not None:
                self.socket.settimeout(None)

    async def ssl_start(self, certfile: str, keyfile: Optional[str] = None) -> None:
        """
        Asynchronously start an SSL connection.

        :param certfile: The path to the certificate file.
        :param keyfile: The path to the key file (optional).
        """
        ssl_context = self.create_ssl_context(certfile, keyfile=keyfile)
        try:
            if self.reader is None:
                # Socket not yet bound to the event loop — create SSL-aware streams directly.
                self.reader, self.writer = await asyncio.open_connection(
                    sock=self.socket,
                    ssl=ssl_context,
                    server_hostname="",
                    ssl_handshake_timeout=DEFAULT_SSL_HANDSHAKE_TIMEOUT,
                )
            else:
                # Socket already registered with the event loop — upgrade in-place.
                # Python 3.11+ provides StreamWriter.start_tls(); older versions
                # need loop.start_tls() and a new StreamWriter bound to the TLS transport.
                await self._ensure_started()
                if hasattr(self.writer, "start_tls"):
                    await asyncio.wait_for(
                        self.writer.start_tls(
                            sslcontext=ssl_context,
                            server_hostname="",
                            ssl_handshake_timeout=DEFAULT_SSL_HANDSHAKE_TIMEOUT,
                        ),
                        timeout=DEFAULT_SSL_HANDSHAKE_TIMEOUT,
                    )
                else:
                    loop = asyncio.get_running_loop()
                    protocol = self.writer._protocol  # type: ignore[attr-defined]
                    tls_transport = await asyncio.wait_for(
                        loop.start_tls(
                            self.writer.transport,
                            protocol,
                            ssl_context,
                            server_side=False,
                            server_hostname="",
                            ssl_handshake_timeout=DEFAULT_SSL_HANDSHAKE_TIMEOUT,
                        ),
                        timeout=DEFAULT_SSL_HANDSHAKE_TIMEOUT,
                    )
                    self.writer = asyncio.StreamWriter(tls_transport, protocol, self.reader, loop)
        except OSError as e:
            raise ConnectionTerminatedError() from e

    async def start(self) -> None:
        """Asynchronously initialize stream reader/writer for the socket."""
        if self.reader is not None and self.writer is not None:
            return
        if self.socket is None:
            raise ConnectionTerminatedError()
        self.reader, self.writer = await asyncio.open_connection(sock=self.socket)

    async def __aenter__(self) -> "ServiceConnection":
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.close()

    def shell(self) -> None:
        """Start an interactive shell."""
        start_ipython_shell(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.Terminal256Formatter(style="native")),
            user_ns={
                "client": self,
            },
        )

    def read(self, size: int) -> bytes:
        """
        Read data from the socket.

        :param size: The amount of data to read.
        :return: The read data.
        """
        result = self.recvall_sync(size)
        self._offset += size
        return result

    def write(self, data: bytes) -> None:
        """
        Write data to the socket.

        :param data: The data to write.
        """
        self.sendall_sync(data)
        self._offset += len(data)

    def tell(self) -> int:
        """
        Get the current offset.

        :return: The current offset.
        """
        return self._offset



from pymobiledevice3.services.dvt.instruments import ChannelService
from pymobiledevice3.services.remote_server import MessageAux


class Graphics(ChannelService):
    IDENTIFIER = "com.apple.instruments.server.services.graphics.opengl"

    def __init__(self, dvt):
        super().__init__(dvt)

    async def __aenter__(self):
        channel = await self._channel_ref()
        await channel.startSamplingAtTimeInterval_(MessageAux().append_obj(0.0))
        await channel.receive_plist()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        channel = await self._channel_ref()
        await channel.stopSampling()

    def __aiter__(self):
        while True:
            yield self._read_once()

    async def _read_once(self):
        channel = await self._channel_ref()
        return await channel.receive_plist()



import pytest

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.pcapd import PcapdService


@pytest.mark.asyncio
async def test_sniffing(lockdown: LockdownClient) -> None:
    """
    Test sniffing device traffic.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    async with PcapdService(lockdown) as pcapd:
        packets = []
        async for packet in pcapd.watch(packets_count=2):
            packets.append(packet)
        assert len(packets) == 2

        first_packet_time = packets[0].seconds + (packets[0].microseconds / 1000000)
        second_packet_time = packets[1].seconds + (packets[1].microseconds / 1000000)
        assert first_packet_time < second_packet_time

        assert len(packets[0].data) == packets[0].packet_length
        assert packets[0].header_version == 2



import pytest
from packaging.version import Version

from pymobiledevice3.bonjour import browse_mobdev2, browse_remoted, browse_remotepairing
from pymobiledevice3.lockdown import LockdownClient


@pytest.mark.asyncio
async def test_mobdev2(lockdown: LockdownClient) -> None:
    await lockdown.set_enable_wifi_connections(True)
    assert len(await browse_mobdev2()) >= 1


@pytest.mark.asyncio
async def test_remoted(lockdown: LockdownClient) -> None:
    if Version(lockdown.product_version) < Version("16.0"):
        pytest.skip("iOS < 16.0")
    assert len(await browse_remoted()) >= 1


@pytest.mark.asyncio
async def test_remotepairing(lockdown: LockdownClient) -> None:
    if Version(lockdown.product_version) < Version("17.0"):
        pytest.skip("iOS < 17.0")
    assert len(await browse_remotepairing()) >= 1



#!/usr/bin/env python3
from typing import Optional

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService


class CompanionProxyService(LockdownService):
    SERVICE_NAME = "com.apple.companion_proxy"
    RSD_SERVICE_NAME = "com.apple.companion_proxy.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    async def list(self):
        service = await self.lockdown.start_lockdown_service(self.service_name)
        return (await service.send_recv_plist({"Command": "GetDeviceRegistry"})).get("PairedDevicesArray", [])

    async def listen_for_devices(self):
        service = await self.lockdown.start_lockdown_service(self.service_name)
        await service.send_plist({"Command": "StartListeningForDevices"})
        while True:
            yield await service.recv_plist()

    async def get_value(self, udid: str, key: str):
        service = await self.lockdown.start_lockdown_service(self.service_name)
        response = await service.send_recv_plist({
            "Command": "GetValueFromRegistry",
            "GetValueGizmoUDIDKey": udid,
            "GetValueKeyKey": key,
        })

        value = response.get("RetrievedValueDictionary")
        if value is not None:
            return value

        error = response.get("Error")
        raise PyMobileDevice3Exception(error)

    async def start_forwarding_service_port(
        self, remote_port: int, service_name: Optional[str] = None, options: Optional[dict] = None
    ):
        service = await self.lockdown.start_lockdown_service(self.service_name)

        request = {
            "Command": "StartForwardingServicePort",
            "GizmoRemotePortNumber": remote_port,
            "IsServiceLowPriority": False,
            "PreferWifi": False,
        }

        if service_name is not None:
            request["ForwardedServiceName"] = service_name

        if options is not None:
            request.update(options)

        return (await service.send_recv_plist(request)).get("CompanionProxyServicePort")

    async def stop_forwarding_service_port(self, remote_port: int):
        service = await self.lockdown.start_lockdown_service(self.service_name)

        request = {"Command": "StopForwardingServicePort", "GizmoRemotePortNumber": remote_port}

        return await service.send_recv_plist(request)



import pytest
from bpylist2 import archiver

from pymobiledevice3.exceptions import DvtException, ExtractingStackshotError, InvalidServiceError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.core_profile_session_tap import (
    CoreProfileSessionTap,
)


@pytest.mark.asyncio
async def test_stackshot(service_provider: LockdownServiceProvider) -> None:
    """
    Test getting stackshot.
    """
    try:
        async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
            time_config = await CoreProfileSessionTap.get_time_config(dvt)
            async with CoreProfileSessionTap(dvt, time_config) as tap:
                data = await tap.get_stackshot()
    except InvalidServiceError:
        pytest.skip("Skipping DVT-based test since the service isn't accessible")

    assert "Darwin Kernel" in data["osversion"]
    # Constant kernel task data.
    assert data["task_snapshots"][0]["task_snapshot"]["ts_pid"] == 0
    assert data["task_snapshots"][0]["task_snapshot"]["ts_p_comm"] == "kernel_task"


@pytest.mark.asyncio
async def test_stackshot_raises_on_start_failure_notice() -> None:
    class FakeChannel:
        async def receive_message(self):
            return archiver.archive({
                "k": 8,
                "status": 2147483648,
                "notice": "Failed to start the recording: _lockKPerf: could not lock kperf.",
            })

    tap = CoreProfileSessionTap(None, {})
    tap.channel = FakeChannel()

    with pytest.raises(DvtException, match="Failed to start the recording"):
        await tap.get_stackshot(timeout=0.1)


@pytest.mark.asyncio
async def test_stackshot_times_out_without_stackshot_data() -> None:
    class FakeChannel:
        async def receive_message(self):
            return archiver.archive({"k": 7, "sm": {}})

    tap = CoreProfileSessionTap(None, {})
    tap.channel = FakeChannel()

    with pytest.raises(ExtractingStackshotError, match="timed out waiting for stackshot data"):
        await tap.get_stackshot(timeout=0.01)



from typing import Optional

import requests

from pymobiledevice3.exceptions import TunneldConnectionError
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService

TUNNELD_DEFAULT_ADDRESS = ("127.0.0.1", 49151)


async def get_tunneld_devices(
    tunneld_address: tuple[str, int] = TUNNELD_DEFAULT_ADDRESS,
) -> list[RemoteServiceDiscoveryService]:
    tunnels = _list_tunnels(tunneld_address)
    return await _create_rsds_from_tunnels(tunnels)


async def get_tunneld_device_by_udid(
    udid: str, tunneld_address: tuple[str, int] = TUNNELD_DEFAULT_ADDRESS
) -> Optional[RemoteServiceDiscoveryService]:
    tunnels = _list_tunnels(tunneld_address)
    if udid not in tunnels:
        return None
    rsds = await _create_rsds_from_tunnels({udid: tunnels[udid]})
    return rsds[0]


def _list_tunnels(tunneld_address: tuple[str, int] = TUNNELD_DEFAULT_ADDRESS) -> dict[str, list[dict]]:
    try:
        # Get the list of tunnels from the specified address
        resp = requests.get(f"http://{tunneld_address[0]}:{tunneld_address[1]}")
        tunnels = resp.json()
    except requests.exceptions.ConnectionError as e:
        raise TunneldConnectionError() from e
    return tunnels


async def _create_rsds_from_tunnels(tunnels: dict[str, list[dict]]) -> list[RemoteServiceDiscoveryService]:
    rsds = []
    for _udid, details in tunnels.items():
        for tunnel_details in details:
            rsd = RemoteServiceDiscoveryService(
                (tunnel_details["tunnel-address"], tunnel_details["tunnel-port"]), name=tunnel_details["interface"]
            )
            try:
                await rsd.connect()
                rsds.append(rsd)
            except (TimeoutError, ConnectionError):
                continue
    return rsds



from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.remote.remote_service import RemoteService
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService


class RestoreService(RemoteService):
    SERVICE_NAME = "com.apple.RestoreRemoteServices.restoreserviced"

    def __init__(self, lockdown: RemoteServiceDiscoveryService):
        super().__init__(lockdown, self.SERVICE_NAME)

    async def delay_recovery_image(self) -> None:
        """
        Set `delay-recovery-image` on devices of ProductType 0x1677b394. Otherwise, fail
        """
        await self.validate_command("delayrecoveryimage")

    async def enter_recovery(self) -> None:
        """Enter recovery"""
        await self.validate_command("recovery")

    async def reboot(self) -> None:
        """Reboot device"""
        await self.validate_command("reboot")

    async def get_preflightinfo(self) -> dict:
        """Get preflight info"""
        return await self.service.send_receive_request({"command": "getpreflightinfo"})

    async def get_nonces(self) -> dict:
        """Get ApNonce and SEPNonce"""
        return await self.service.send_receive_request({"command": "getnonces"})

    async def get_app_parameters(self) -> dict:
        return await self.validate_command("getappparameters")

    async def restore_lang(self, language: str) -> dict:
        return await self.service.send_receive_request({"command": "restorelang", "argument": language})

    async def validate_command(self, command: str) -> dict:
        """Execute command and validate result is `success`"""
        response = await self.service.send_receive_request({"command": command})
        if response.get("result") != "success":
            raise PyMobileDevice3Exception(f"request command: {command} failed with error: {response}")
        return response



#!/usr/bin/env python3
import asyncio
import plistlib
import uuid
from asyncio import IncompleteReadError
from contextlib import asynccontextmanager, suppress
from datetime import datetime
from pathlib import Path
from typing import Union

from pymobiledevice3.exceptions import (
    AfcException,
    AfcFileNotFoundError,
    LockdownError,
    PyMobileDevice3Exception,
)
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.afc import AFC_LOCK_EX, AFC_LOCK_UN, AfcError, AfcService
from pymobiledevice3.services.device_link import DeviceLink
from pymobiledevice3.services.installation_proxy import InstallationProxyService
from pymobiledevice3.services.lockdown_service import LockdownService
from pymobiledevice3.services.notification_proxy import NotificationProxyService
from pymobiledevice3.services.springboard import SpringBoardServicesService

SUPPORTED_VERSIONS = [2.0, 2.1]
ITUNES_FILES = [
    "ApertureAlbumPrefs",
    "IC-Info.sidb",
    "IC-Info.sidv",
    "PhotosFolderAlbums",
    "PhotosFolderName",
    "PhotosFolderPrefs",
    "VoiceMemos.plist",
    "iPhotoAlbumPrefs",
    "iTunesApplicationIDs",
    "iTunesPrefs",
    "iTunesPrefs.plist",
]
NP_SYNC_WILL_START = "com.apple.itunes-mobdev.syncWillStart"
NP_SYNC_DID_START = "com.apple.itunes-mobdev.syncDidStart"
NP_SYNC_LOCK_REQUEST = "com.apple.itunes-mobdev.syncLockRequest"
NP_SYNC_DID_FINISH = "com.apple.itunes-mobdev.syncDidFinish"


class Mobilebackup2Service(LockdownService):
    SERVICE_NAME = "com.apple.mobilebackup2"
    RSD_SERVICE_NAME = "com.apple.mobilebackup2.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME, include_escrow_bag=True)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME, include_escrow_bag=True)

    @property
    def will_encrypt(self) -> bool:
        try:
            domain_values = self.lockdown.all_values.get("com.apple.mobile.backup", {})
            return bool(domain_values.get("WillEncrypt", False))
        except LockdownError:
            return False

    async def backup(
        self, full: bool = True, backup_directory: Union[str, Path] = ".", progress_callback=lambda x: None
    ) -> None:
        """
        Backup a device.
        :param full: Whether to do a full backup. If full is True, any previous backup attempts will be discarded.
        :param backup_directory: Directory to write backup to.
        :param progress_callback: Function to be called as the backup progresses.
        The function shall receive the percentage as a parameter.
        """
        backup_directory = Path(backup_directory)
        device_directory = backup_directory / self.lockdown.udid
        device_directory.mkdir(exist_ok=True, mode=0o755, parents=True)

        async with (
            self.device_link(backup_directory) as dl,
            NotificationProxyService(self.lockdown) as notification_proxy,
            AfcService(self.lockdown) as afc,
            self._backup_lock(afc, notification_proxy),
        ):
            # Initialize Info.plist
            info_plist = await self.init_mobile_backup_factory_info(afc)
            with open(device_directory / "Info.plist", "wb") as fd:
                plistlib.dump(info_plist, fd)

            # Initialize Status.plist file if doesn't exist.
            status_path = device_directory / "Status.plist"
            current_date = datetime.now()
            current_date = current_date.replace(tzinfo=None)
            if full or not status_path.exists():
                with open(device_directory / "Status.plist", "wb") as fd:
                    plistlib.dump(
                        {
                            "BackupState": "new",
                            "Date": current_date,
                            "IsFullBackup": full,
                            "Version": "3.3",
                            "SnapshotState": "finished",
                            "UUID": str(uuid.uuid4()).upper(),
                        },
                        fd,
                        fmt=plistlib.FMT_BINARY,
                    )

            # Create Manifest.plist if doesn't exist.
            manifest_path = device_directory / "Manifest.plist"
            if full:
                manifest_path.unlink(missing_ok=True)
            (device_directory / "Manifest.plist").touch()

            await dl.send_process_message({"MessageName": "Backup", "TargetIdentifier": self.lockdown.udid})
            await dl.dl_loop(progress_callback)

    async def restore(
        self,
        backup_directory=".",
        system: bool = False,
        reboot: bool = True,
        copy: bool = True,
        settings: bool = True,
        remove: bool = False,
        password: str = "",
        source: str = "",
        progress_callback=lambda x: None,
        skip_apps: bool = False,
    ):
        """
        Restore a previous backup to the device.
        :param backup_directory: Path of the backup directory.
        :param system: Whether to restore system files.
        :param reboot: Reboot the device when done.
        :param copy: Create a copy of backup folder before restoring.
        :param settings: Restore device settings.
        :param remove: Remove items which aren't being restored.
        :param password: Password of the backup if it is encrypted.
        :param source: Identifier of device to restore its backup.
        :param progress_callback: Function to be called as the backup progresses.
        :param skip_apps: Do not trigger re-installation of apps after restore.
        The function shall receive the current percentage of the progress as a parameter.
        """
        backup_directory = Path(backup_directory)
        source = source if source else self.lockdown.udid
        self._assert_backup_exists(backup_directory, source)

        async with (
            self.device_link(backup_directory) as dl,
            NotificationProxyService(self.lockdown) as notification_proxy,
            AfcService(self.lockdown) as afc,
            self._backup_lock(afc, notification_proxy),
        ):
            manifest_plist_path = backup_directory / source / "Manifest.plist"
            with open(manifest_plist_path, "rb") as fd:
                manifest = plistlib.load(fd)
            is_encrypted = manifest.get("IsEncrypted", False)
            options = {
                "RestoreShouldReboot": reboot,
                "RestoreDontCopyBackup": not copy,
                "RestorePreserveSettings": settings,
                "RestoreSystemFiles": system,
                "RemoveItemsNotRestored": remove,
            }
            if is_encrypted:
                if password:
                    options["Password"] = password
                else:
                    self.logger.error("Backup is encrypted, please supply password.")
                    return
            await dl.send_process_message({
                "MessageName": "Restore",
                "TargetIdentifier": self.lockdown.udid,
                "SourceIdentifier": source,
                "Options": options,
            })

            if not skip_apps:
                # Write /iTunesRestore/RestoreApplications.plist so that the device will start
                # restoring applications once the rest of the restore process is finished
                info_plist_path = backup_directory / source / "Info.plist"
                applications = plistlib.loads(info_plist_path.read_bytes()).get("Applications")
                if applications is not None:
                    await afc.makedirs("/iTunesRestore")
                    await afc.set_file_contents(
                        "/iTunesRestore/RestoreApplications.plist", plistlib.dumps(applications)
                    )

            await dl.dl_loop(progress_callback)

    async def info(self, backup_directory=".", source: str = "") -> str:
        """
        Get information about a backup.
        :param backup_directory: Path of the backup directory.
        :param source: Identifier of device to get info about its backup.
        :return: Information about a backup.
        """
        backup_dir = Path(backup_directory)
        self._assert_backup_exists(backup_dir, source if source else self.lockdown.udid)
        async with self.device_link(backup_dir) as dl:
            message = {"MessageName": "Info", "TargetIdentifier": self.lockdown.udid}
            if source:
                message["SourceIdentifier"] = source
            await dl.send_process_message(message)
            result = await dl.dl_loop()
        return result

    async def list(self, backup_directory=".", source: str = "") -> str:
        """
        List the files in the last backup.
        :param backup_directory: Path of the backup directory.
        :param source: Identifier of device to list its backup data.
        :return: List of files and additional data about each file, all in a CSV format.
        """
        backup_dir = Path(backup_directory)
        source = source if source else self.lockdown.udid
        self._assert_backup_exists(backup_dir, source)
        async with self.device_link(backup_dir) as dl:
            await dl.send_process_message({
                "MessageName": "List",
                "TargetIdentifier": self.lockdown.udid,
                "SourceIdentifier": source,
            })
            result = await dl.dl_loop()
        return result

    async def unback(self, backup_directory=".", password: str = "", source: str = "") -> None:
        """
        Unpack a complete backup to its device hierarchy.
        :param backup_directory: Path of the backup directory.
        :param password: Password of the backup if it is encrypted.
        :param source: Identifier of device to unpack its backup.
        """
        backup_dir = Path(backup_directory)
        self._assert_backup_exists(backup_dir, source if source else self.lockdown.udid)
        async with self.device_link(backup_dir) as dl:
            message = {"MessageName": "Unback", "TargetIdentifier": self.lockdown.udid}
            if source:
                message["SourceIdentifier"] = source
            if password:
                message["Password"] = password
            await dl.send_process_message(message)
            await dl.dl_loop()

    async def extract(
        self, domain_name: str, relative_path: str, backup_directory=".", password: str = "", source: str = ""
    ) -> None:
        """
        Extract a file from a previous backup.
        :param domain_name: File's domain name, e.g., SystemPreferencesDomain or HomeDomain.
        :param relative_path: File path.
        :param backup_directory: Path of the backup directory.
        :param password: Password of the last backup if it is encrypted.
        :param source: Identifier of device to extract file from its backup.
        """
        backup_dir = Path(backup_directory)
        self._assert_backup_exists(backup_dir, source if source else self.lockdown.udid)
        async with self.device_link(backup_dir) as dl:
            message = {
                "MessageName": "Extract",
                "TargetIdentifier": self.lockdown.udid,
                "DomainName": domain_name,
                "RelativePath": relative_path,
            }
            if source:
                message["SourceIdentifier"] = source
            if password:
                message["Password"] = password
            await dl.send_process_message(message)
            await dl.dl_loop()

    async def change_password(self, backup_directory=".", old: str = "", new: str = "") -> None:
        """
        Change backup password.
        :param backup_directory: Backups directory.
        :param old: Previous password. Omit when enabling backup encryption.
        :param new: New password. Omit when disabling backup encryption.
        """
        async with self.device_link(Path(backup_directory)) as dl:
            message = {"MessageName": "ChangePassword", "TargetIdentifier": self.lockdown.udid}
            if old:
                message["OldPassword"] = old
            if new:
                message["NewPassword"] = new
            await dl.send_process_message(message)
            await dl.dl_loop()

    async def erase_device(self, backup_directory=".") -> None:
        """
        Erase the device.
        """
        with suppress(IncompleteReadError):
            async with self.device_link(Path(backup_directory)) as dl:
                await dl.send_process_message({"MessageName": "EraseDevice", "TargetIdentifier": self.lockdown.udid})
                await dl.dl_loop()

    async def version_exchange(self, dl: DeviceLink, local_versions=None) -> None:
        """
        Exchange versions with the device and assert that the device supports our version of the protocol.
        :param dl: Initialized device link.
        :param local_versions: versions supported by us.
        """
        if local_versions is None:
            local_versions = SUPPORTED_VERSIONS
        await dl.send_process_message({
            "MessageName": "Hello",
            "SupportedProtocolVersions": local_versions,
        })
        reply = await dl.receive_message()
        assert reply[0] == "DLMessageProcessMessage" and reply[1]["ErrorCode"] == 0
        assert reply[1]["ProtocolVersion"] in local_versions

    async def init_mobile_backup_factory_info(self, afc: AfcService):
        async with InstallationProxyService(self.lockdown) as ip, SpringBoardServicesService(self.lockdown) as sbs:
            root_node = self.lockdown.all_values
            itunes_settings = self.lockdown.all_values.get("com.apple.iTunes", {})
            min_itunes_version = self.lockdown.all_values.get("com.apple.mobile.iTunes", {}).get("MinITunesVersion")
            if min_itunes_version is None:
                # iPadOS may not contain this value. See:
                # https://github.com/doronz88/pymobiledevice3/issues/1332
                min_itunes_version = "10.0.1"
            app_dict = {}
            installed_apps = []
            apps = await ip.browse(
                options={"ApplicationType": "User"},
                attributes=["CFBundleIdentifier", "ApplicationSINF", "iTunesMetadata"],
            )
            for app in apps:
                bundle_id = app["CFBundleIdentifier"]
                if bundle_id:
                    installed_apps.append(bundle_id)
                    if app.get("iTunesMetadata", False) and app.get("ApplicationSINF", False):
                        app_dict[bundle_id] = {
                            "ApplicationSINF": app["ApplicationSINF"],
                            "iTunesMetadata": app["iTunesMetadata"],
                            "PlaceholderIcon": await sbs.get_icon_pngdata(bundle_id),
                        }

            files = {}
            for file in ITUNES_FILES:
                try:
                    data_buf = await afc.get_file_contents("/iTunes_Control/iTunes/" + file)
                except AfcFileNotFoundError:
                    pass
                else:
                    files[file] = data_buf

            ret = {
                "iTunes Version": min_itunes_version if min_itunes_version else "10.0.1",
                "iTunes Files": files,
                "Unique Identifier": self.lockdown.udid.upper(),
                "Target Type": "Device",
                "Target Identifier": root_node["UniqueDeviceID"],
                "Serial Number": root_node["SerialNumber"],
                "Product Version": root_node["ProductVersion"],
                "Product Type": root_node["ProductType"],
                "Installed Applications": installed_apps,
                "GUID": uuid.uuid4().bytes,
                "Display Name": root_node["DeviceName"],
                "Device Name": root_node["DeviceName"],
                "Build Version": root_node["BuildVersion"],
                "Applications": app_dict,
            }

            if "IntegratedCircuitCardIdentity" in root_node:
                ret["ICCID"] = root_node["IntegratedCircuitCardIdentity"]
            if "InternationalMobileEquipmentIdentity" in root_node:
                ret["IMEI"] = root_node["InternationalMobileEquipmentIdentity"]
            if "MobileEquipmentIdentifier" in root_node:
                ret["MEID"] = root_node["MobileEquipmentIdentifier"]
            if "PhoneNumber" in root_node:
                ret["Phone Number"] = root_node["PhoneNumber"]

            try:
                data_buf = await afc.get_file_contents("/Books/iBooksData2.plist")
            except AfcFileNotFoundError:
                pass
            else:
                ret["iBooks Data 2"] = data_buf
            if itunes_settings:
                ret["iTunes Settings"] = itunes_settings
            return ret

    @asynccontextmanager
    async def _backup_lock(self, afc, notification_proxy):
        await notification_proxy.notify_post(NP_SYNC_WILL_START)
        lockfile = await afc.fopen("/com.apple.itunes.lock_sync", "r+")
        if lockfile:
            await notification_proxy.notify_post(NP_SYNC_LOCK_REQUEST)
            for _ in range(50):
                try:
                    await afc.lock(lockfile, AFC_LOCK_EX)
                except AfcException as e:
                    if e.status == AfcError.OP_WOULD_BLOCK:
                        await asyncio.sleep(0.2)
                    else:
                        await afc.fclose(lockfile)
                        raise
                else:
                    await notification_proxy.notify_post(NP_SYNC_DID_START)
                    break
            else:  # No break, lock failed.
                await afc.fclose(lockfile)
                raise PyMobileDevice3Exception("Failed to lock itunes sync file")
        try:
            yield
        finally:
            await afc.lock(lockfile, AFC_LOCK_UN)
            await afc.fclose(lockfile)
            await notification_proxy.notify_post(NP_SYNC_DID_FINISH)

    @staticmethod
    def _assert_backup_exists(backup_directory: Path, identifier: str):
        device_directory = backup_directory / identifier
        assert (device_directory / "Info.plist").exists()
        assert (device_directory / "Manifest.plist").exists()
        assert (device_directory / "Status.plist").exists()

    @asynccontextmanager
    async def device_link(self, backup_directory):
        dl = DeviceLink(self.service, backup_directory)
        await dl.version_exchange()
        await self.version_exchange(dl)
        try:
            yield dl
        finally:
            await dl.disconnect()



import contextlib
import platform
from collections.abc import Generator
from typing import Optional

import psutil

from pymobiledevice3.bonjour import DEFAULT_BONJOUR_TIMEOUT, browse_remoted
from pymobiledevice3.exceptions import AccessDeniedError
from pymobiledevice3.remote.remote_service_discovery import RSD_PORT, RemoteServiceDiscoveryService

REMOTED_PATH = "/usr/libexec/remoted"


async def get_rsds(
    bonjour_timeout: float = DEFAULT_BONJOUR_TIMEOUT, udid: Optional[str] = None
) -> list[RemoteServiceDiscoveryService]:
    result = []
    with stop_remoted():
        for answer in await browse_remoted(timeout=bonjour_timeout):
            for address in answer.addresses:
                rsd = RemoteServiceDiscoveryService((address.full_ip, RSD_PORT))
                try:
                    await rsd.connect()
                except ConnectionRefusedError:
                    continue
                except OSError:
                    continue
                if udid is None or rsd.udid == udid:
                    result.append(rsd)
                else:
                    await rsd.close()
    return result


def get_remoted_process() -> psutil.Process:
    for process in psutil.process_iter():
        if process.pid == 0:
            # skip kernel task
            continue
        try:
            if process.exe() == REMOTED_PATH:
                return process
        except (psutil.ZombieProcess, psutil.NoSuchProcess):
            continue


def stop_remoted_if_required() -> None:
    if platform.system() != "Darwin":
        # only Darwin systems require it
        return

    remoted = get_remoted_process()
    if remoted is None:
        return
    if remoted.status() == "stopped":
        # process already stopped, we don't need to do anything
        return

    try:
        remoted.suspend()
    except psutil.AccessDenied as e:
        raise AccessDeniedError() from e


def resume_remoted_if_required() -> None:
    if platform.system() != "Darwin":
        # only Darwin systems require it
        return

    remoted = get_remoted_process()
    if remoted is None:
        return
    if remoted.status() == "running":
        # process already running, we don't need to do anything
        return

    try:
        remoted.resume()
    except psutil.AccessDenied as e:
        raise AccessDeniedError() from e


@contextlib.contextmanager
def stop_remoted() -> Generator[None, None, None]:
    stop_remoted_if_required()
    try:
        yield
    finally:
        resume_remoted_if_required()
